import json
from typing import Any, Dict, List

import pytest
from aioresponses import aioresponses

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_result import (
    GetBatchResult,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)


async def test_get_batch_result_streaming_async() -> None:
    # Mock server URL
    server_url = "http://mockserver.com/fhir"

    # Mock parameters
    parameters = get_fhir_receiver_parameters()

    # Mock response data
    mock_response_data = {
        "resourceType": "Bundle",
        "type": "searchset",
        "total": 1,
        "entry": [
            {
                "resource": {
                    "id": "1",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            }
        ],
    }

    # Expected result
    expected_result: GetBatchResult = GetBatchResult(
        resources=[
            json.dumps(
                {
                    "id": "1",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            )
        ],
        errors=[],
    )

    # Mock the HTTP request
    with aioresponses() as m:
        m.get(f"{server_url}/Patient", payload=mock_response_data)
        m.get(f"{server_url}/Patient?id:above=1", payload=[])

        # Call the function
        results: List[Dict[str, Any]] = []
        async for result in FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url=server_url,
        ):
            results.append(result)

        # Assert the results
        assert len(results) == 2
        assert results[0] == expected_result.to_dict()


async def test_get_batch_result_streaming_async_single_result() -> None:
    # Mock server URL
    server_url = "http://mockserver.com/fhir"

    # Mock parameters
    parameters = get_fhir_receiver_parameters()

    # Mock response data
    mock_response_data = {
        "resourceType": "Bundle",
        "type": "searchset",
        "total": 1,
        "entry": [
            {
                "resource": {
                    "id": "1",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            }
        ],
    }

    # Expected result
    expected_result = GetBatchResult(
        resources=[
            json.dumps(
                {
                    "id": "1",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            )
        ],
        errors=[],
    )

    # Mock the HTTP request
    with aioresponses() as m:
        m.get(f"{server_url}/Patient", payload=mock_response_data)
        m.get(f"{server_url}/Patient?id:above=1", payload=[])

        # Call the function
        results: List[Dict[str, Any]] = []
        async for result in FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url=server_url,
        ):
            results.append(result)

        # Assert the results
        assert len(results) == 2
        assert results[0] == expected_result.to_dict()


@pytest.mark.asyncio
async def test_get_batch_result_streaming_async_multiple_results() -> None:
    # Mock server URL
    server_url = "http://mockserver.com/fhir"

    # Mock parameters
    parameters = get_fhir_receiver_parameters()

    # Mock response data
    mock_response_data = {
        "resourceType": "Bundle",
        "type": "searchset",
        "total": 2,
        "entry": [
            {
                "resource": {
                    "id": "1",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            },
            {
                "resource": {
                    "id": "2",
                    "resourceType": "Patient",
                    # Add other necessary fields here
                }
            },
        ],
    }

    # Expected result
    expected_results = [
        GetBatchResult(
            resources=[
                json.dumps(
                    {
                        "id": "1",
                        "resourceType": "Patient",
                    }
                ),
                json.dumps(
                    {
                        "id": "2",
                        "resourceType": "Patient",
                    }
                ),
            ],
            errors=[],
        )
    ]

    # Mock the HTTP request
    with aioresponses() as m:
        m.get(f"{server_url}/Patient", payload=mock_response_data)
        m.get(f"{server_url}/Patient?id:above=2", payload=[])

        # Call the function
        results: List[Dict[str, Any]] = []
        async for result in FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url=server_url,
        ):
            results.append(result)

        # Assert the results
        assert len(results) == 2
        # Asserting the first dictionary only, since the second one is empty.
        assert results[0] == expected_results[0].to_dict()


@pytest.mark.asyncio
async def test_get_batch_result_streaming_async_no_results() -> None:
    # Mock server URL
    server_url = "http://mockserver.com/fhir"

    # Mock parameters
    parameters = get_fhir_receiver_parameters()
    # Mock response data
    mock_response_data = {
        "resourceType": "Bundle",
        "type": "searchset",
        "total": 0,
        "entry": [],
    }

    # Mock the HTTP request
    with aioresponses() as m:
        m.get(f"{server_url}/Patient", payload=mock_response_data)

        # Call the function
        results: List[Dict[str, Any]] = []
        async for result in FhirReceiverProcessor.get_batch_result_streaming_async(
            last_updated_after=None,
            last_updated_before=None,
            parameters=parameters,
            server_url=server_url,
        ):
            results.append(result)

        # Assert the results
        assert len(results) == 1
        assert results[0]["resources"] == []
        assert results[0]["errors"] == []



import os
from typing import Tuple, Optional

from opensearchpy import OpenSearch, AsyncOpenSearch

from furl import furl

from spark_pipeline_framework.utilities.aws.config import get_ssm_config
from botocore.exceptions import ClientError
from spark_pipeline_framework.logger.yarn_logger import get_logger


class ElasticSearchConnection:
    def get_elastic_search_client(self) -> OpenSearch:
        env_name = os.environ.get("ENV", "local")
        es_url = self._get_connection_string()
        client = OpenSearch(
            hosts=es_url,
            use_ssl=env_name != "local",
            verify_certs=env_name != "local",
        )
        return client

    def get_elastic_search_async_client(self, *, timeout: int = 60) -> AsyncOpenSearch:
        """
        Create an async OpenSearch client

        :param timeout: timeout in seconds for calls made by this client
        :return: async OpenSearch client
        """
        env_name = os.environ.get("ENV", "local")
        es_url = self._get_connection_string()
        client = AsyncOpenSearch(
            hosts=es_url,
            use_ssl=env_name != "local",
            verify_certs=env_name != "local",
            timeout=timeout,  # timeout in seconds
        )
        return client

    @staticmethod
    def get_elastic_search_host() -> str:
        host: str = os.environ["ELASTICSEARCH_HOST"]
        return host

    @staticmethod
    def _get_ssm_path(env_name: str) -> str:
        return f"/{env_name}/helix/elasticsearch/"

    @staticmethod
    def _get_connection_string() -> str:
        env_name = os.environ.get("ENV", "local")
        host: furl = furl(ElasticSearchConnection.get_elastic_search_host())
        port: str = os.environ.get("ELASTICSEARCH_PORT", host.port)
        user, password = ElasticSearchConnection._get_user_and_password(
            env_name=env_name
        )
        host.set(username=user, password=password, port=port)
        return str(host.url)

    @staticmethod
    def _get_user_and_password(env_name: str) -> Tuple[Optional[str], Optional[str]]:
        logger = get_logger(__name__)
        if env_name == "local":
            return "admin", "admin"
        username: Optional[str] = os.environ.get("ELASTICSEARCH_USER")
        password: Optional[str] = os.environ.get("ELASTICSEARCH_PASSWORD")
        # if password is set an environment variable then use that
        if password:
            return username, password
        # else get from SSM
        path: str = ElasticSearchConnection._get_ssm_path(env_name=env_name)
        try:
            db_config = get_ssm_config(path=path)
        except ClientError:
            bwell_env = os.environ.get("BWELL_ENV", "local")
            bwell_path: str = ElasticSearchConnection._get_ssm_path(env_name=bwell_env)
            logger.info(
                f"Unable to find SSM path via ENV: {path}; trying to use BWELL_ENV instead: {bwell_path}"
            )
            db_config = get_ssm_config(path=bwell_path)
            path = bwell_path

        username = db_config[f"{path}username"]
        password = db_config[f"{path}password"]
        return username, password



import os
from typing import Dict, Any, Optional, List, Tuple, Iterable

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.rdd import RDD
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType, StructField, StringType, Row
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_parse_bundles import (
    extract_resource_from_json,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class FixFhirBundleTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        input_path: str,
        output_path: str,
        resources_to_extract: List[str] = [],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        assert isinstance(input_path, str)
        self.input_path: Param[str] = Param(self, "input_path", "")
        self._setDefault(input_path=None)
        assert isinstance(output_path, str)
        self.output_path: Param[str] = Param(self, "output_path", "")
        self._setDefault(output_path=None)
        self.resources_to_extract: Param[List[str]] = Param(
            self, "resources_to_extract", ""
        )
        assert isinstance(resources_to_extract, list)
        self._setDefault(resources_to_extract=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        # init
        spark_session: SparkSession = df.sparkSession
        input_path: str = self.getInputPath()
        output_path: str = self.getOutputPath()
        resources_to_extract: List[str] = self.getResourcesToExtract()

        self.logger.info(f"Reading from {input_path}")
        # separate bundles.
        # This returns one row per file where the key is the name of the file and the value is the contents
        whole_text_rdd: RDD[Tuple[str, str]] = (
            spark_session.sparkContext.wholeTextFiles(input_path)
        )

        resource_list_rdd: RDD[Iterable[Row]] = whole_text_rdd.map(
            lambda x: extract_resource_from_json(
                x[0], x[1], resources_to_extract
            )  # x[0] is file name, x[1] is contents
        )

        resource_rdd: RDD[Row] = resource_list_rdd.flatMap(
            lambda x: x
        )  # convert list of Rows to Rows

        resource_df: DataFrame = resource_rdd.toDF(
            schema=StructType(
                [
                    StructField("file_name", StringType()),
                    StructField("resourceType", StringType()),
                    StructField("resource", StringType()),
                ]
            )
        )

        # now remove any OperationOutcome resources which are errors
        operation_outcome_resource_type: str = "OperationOutcome"
        errors_df: DataFrame = resource_df.where(
            col("resourceType") == operation_outcome_resource_type
        )
        if not spark_is_data_frame_empty(errors_df):
            # write out the errors
            output_file_path = os.path.join(
                output_path, operation_outcome_resource_type
            )
            self.logger.info(
                f"Found errors. Writing {operation_outcome_resource_type} to {output_file_path}"
            )
            errors_df.select("resource").write.text(output_file_path)

        # process resources that are not errors
        resource_df = resource_df.where(
            col("resourceType") != operation_outcome_resource_type
        )

        # now separately write out each resource we're interested in
        if not resources_to_extract:
            resources_to_extract = [
                r.resourceType
                for r in resource_df.select("resourceType").distinct().collect()
            ]

        self.logger.info(
            f"Looking for resources {','.join(resources_to_extract)} in {input_path}"
        )
        for resource_to_extract in resources_to_extract:
            extract_df: DataFrame = resource_df.where(
                resource_df["resourceType"] == resource_to_extract
            )
            if not spark_is_data_frame_empty(extract_df):
                output_file_path = os.path.join(output_path, resource_to_extract)
                self.logger.info(f"Writing {resource_to_extract} to {output_file_path}")
                extract_df.select("resource").write.text(output_file_path)
            else:
                self.logger.warning(f"No {resource_to_extract} resources found")

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getInputPath(self) -> str:
        return self.getOrDefault(self.input_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputPath(self) -> str:
        return self.getOrDefault(self.output_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResourcesToExtract(self) -> List[str]:
        return self.getOrDefault(self.resources_to_extract)



import json
from typing import Any, Optional

import requests
from requests.adapters import HTTPAdapter

from spark_pipeline_framework.utilities.slack.base_slack_client import BaseSlackClient


class SlackClient(BaseSlackClient):
    """
    Implements a basic Slack client in http to post messages to a channel
    Inspired by https://keestalkstech.com/2019/10/simple-python-code-to-send-message-to-slack-channel-without-packages/
    """

    def __init__(self, slack_token: str, channel: str, bot_user_name: str) -> None:
        """
        Client for sending messages to Slack

        :param slack_token: token to auth with Slack
        :param channel: channel to post message into
        :param bot_user_name: user name to use when posting messages
        """
        self.slack_token: str = slack_token
        self.slack_channel: str = channel
        self.slack_icon_url: str = "https://www.flaticon.com/free-icon/freepik_23317"
        self.slack_user_name: str = bot_user_name
        self.slack_thread: Optional[str] = None

    def post_message_to_slack(
        self,
        text: str,
        blocks: Any = None,
        use_conversation_threads: bool = True,
        slack_thread: Optional[str] = None,
    ) -> Any:
        """
        Posts a message to Slack
        :param text: message to post.  Can be markdown.
        :param blocks: (Optional) blocks to post
        :param use_conversation_threads: whether to send messages as reply to first message
        :return: response from Slack
        """
        if not text:
            return

        headers = {
            "Content-type": "application/json; charset=utf-8",
            "Authorization": f"Bearer {self.slack_token}",
        }
        adapter = HTTPAdapter()
        http = requests.Session()
        http.mount("https://", adapter)

        post = http.post(
            "https://slack.com/api/chat.postMessage",
            data=json.dumps(
                {
                    "channel": self.slack_channel,
                    "text": text,
                    "icon_url": self.slack_icon_url,
                    "username": self.slack_user_name,
                    "blocks": json.dumps(blocks) if blocks else None,
                    "thread_ts": (
                        slack_thread if slack_thread is not None else self.slack_thread
                    ),
                }
            ),
            headers=headers,
        )
        response_json = post.json()
        if not self.slack_thread and use_conversation_threads:
            if "ok" in response_json and response_json["ok"] is True:
                # read thread id from response
                if "message" in response_json and "ts" in response_json["message"]:
                    self.slack_thread = response_json["message"]["ts"]
        return response_json



from dataclasses import dataclass
from typing import List


@dataclass
class ChunkContainer[TInputColumnDataType]:
    """
    This class is used to store the chunk of input values that will be passed to the Pandas UDF
    """

    """ The input values for the chunk"""
    chunk_input_values: List[TInputColumnDataType]

    """ The index of the chunk"""
    chunk_index: int

    """ The index of the first input value in the chunk"""
    begin_chunk_input_values_index: int

    """ The index of the last input value in the chunk"""
    end_chunk_input_values_index: int



from collections import namedtuple
from typing import Any, Dict

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)


class StandardizedAddress(RawAddress):
    """
    The address standardized by a vendor
    """

    _Address = namedtuple(
        "_Address",
        [
            "address_id",
            "line1",
            "line2",
            "city",
            "county",
            "state",
            "zipcode",
            "country",
            "latitude",
            "longitude",
            "formatted_address",
            "standardize_vendor",
        ],
    )

    def __init__(
        self,
        address_id: str,
        line1: str,
        city: str = "",
        state: str = "",
        zipcode: str = "",
        country: str = "US",
        latitude: str = "",
        longitude: str = "",
        line2: str = "",
        county: str = "",
        formatted_address: str = "",
        standardize_vendor: str = "melissa",
    ):
        self.address: Any = self._Address(
            address_id,
            line1,
            line2,
            city,
            county,
            state,
            zipcode,
            country,
            latitude,
            longitude,
            formatted_address,
            standardize_vendor,
        )

    def to_str(self) -> str:
        if self.address.formatted_address:
            return str(self.address.formatted_address)
        else:
            return super(StandardizedAddress, self).to_str()

    def to_dict(self) -> Dict[str, str]:
        return dict(self.address._asdict())

    @classmethod
    def from_dict(cls, address_dict: Dict[str, str]) -> "StandardizedAddress":
        return cls(
            address_id=address_dict["address_id"],
            line1=address_dict["line1"],
            line2=address_dict["line2"],
            city=address_dict["city"],
            state=address_dict["state"],
            zipcode=address_dict["zipcode"],
            country=address_dict["country"],
            latitude=address_dict["latitude"],
            longitude=address_dict["longitude"],
            formatted_address=address_dict["formatted_address"],
            standardize_vendor=address_dict["standardize_vendor"],
        )

    @classmethod
    def from_raw_address(cls, raw_address: RawAddress) -> "StandardizedAddress":
        return cls(
            address_id=raw_address.get_id(),
            line1=raw_address.address.line1,
            line2=raw_address.address.line2,
            city=raw_address.address.city,
            state=raw_address.address.state,
            zipcode=raw_address.address.zipcode,
            country=raw_address.address.country,
        )



import json
from typing import Any, Dict, Iterable, List, Optional, Callable

from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import (
    to_json,
    struct,
    monotonically_increasing_id,
    when,
    col,
    lit,
)
from pyspark.sql.types import Row, StructType, StructField, StringType
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)

from spark_pipeline_framework.utilities.helix_geolocation.v1.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardize_address import (
    StandardizeAddr,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class AddressStandardization(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        address_column_mapping: Dict[str, str],
        standardizing_vendor: StandardizingVendor,
        cache_handler: CacheHandler,
        func_get_response_path: Callable[[str], str] | None = None,
        geolocation_column_prefix: Optional[str] = "",
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Standardize and geocode addresses in a view using the specified standardizing vendor. Address columns in the view
        should be mapped to the standard address columns like the following example. the values in the dictionary should
        be the column names in the view.
        {
            "line1": "address1",
            "line2": "address2",
            "city": "city",
            "state": "state",
            "zipcode": "zip"
        }

        :param view: view with the address data to standardize
        :param address_column_mapping: mapping of address columns to standardize
        :param standardizing_vendor: vendor class to use for standardization
        :param cache_handler: cache handler to use
        :param func_get_response_path: function to get response path, this should be used if the data in the view is large,
        if it is not specified then the results will be cached
        :param geolocation_column_prefix: prefix for geolocation columns
        :param name: name for the stage in spark
        :param parameters: additional parameters
        :param progress_logger: progress logger
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.address_column_mapping: Param[Dict[str, str]] = Param(
            self, "address_column_mapping", ""
        )
        self._setDefault(address_column_mapping=address_column_mapping)

        self.standardizing_vendor: Param[StandardizingVendor] = Param(
            self, "standardizing_vendor", ""
        )
        self._setDefault(view=standardizing_vendor)

        self.cache_handler: Param[CacheHandler] = Param(self, "cache_handler", "")
        self._setDefault(view=cache_handler)

        self.geolocation_column_prefix: Param[Optional[str]] = Param(
            self, "geolocation_column_prefix", ""
        )
        self._setDefault(geolocation_column_prefix=geolocation_column_prefix)

        self.func_get_response_path: Param[Callable[[str], str]] = Param(
            self, "func_get_response_path", ""
        )
        self._setDefault(func_get_response_path=func_get_response_path)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        try:
            view: str = self.getView()
            func_get_response_path: Callable[[str], str] | None = (
                self.getFuncGetResponsePath()
            )
            address_column_mapping: Dict[str, str] = self.getAddressColumnMapping()
            address_column_mapping_with_id: List[str] = (
                self.get_address_columns_for_view(list(address_column_mapping.values()))
            )
            standardizing_vendor: StandardizingVendor = self.getStandardizingVendor()
            cache_handler: CacheHandler = self.getCacheHandler()
            geolocation_column_prefix: Optional[str] = self.getGeolocationColumnPrefix()
            address_df: DataFrame = (
                df.sql_ctx.table(view)
                .withColumn(
                    # create an id column so we can join back to it once we are done standardizing
                    "address_id",
                    monotonically_increasing_id(),
                )
                .withColumn(
                    # create a new column that contains all the address parts as json string
                    "raw_address",
                    to_json(
                        struct(
                            *[
                                when(col(x).isNotNull(), col(x))
                                .otherwise(lit(""))
                                .alias(x)
                                for x in address_column_mapping_with_id
                            ]
                        )
                    ),
                )
            )

            df.sql_ctx.dropTempTable(view)

            def standardize(rows: Iterable[Row]) -> List[Dict[str, str]]:
                try:
                    register()

                    raw_addresses: List[Dict[str, str]] = [
                        json.loads(r["raw_address"]) for r in rows
                    ]
                    # create raw address objects to send to the standardization module..
                    raw_address_list: List[RawAddress] = [
                        RawAddress(
                            address_id=raw_address["address_id"],
                            line1=raw_address[address_column_mapping["line1"]],
                            line2=raw_address[address_column_mapping["line2"]],
                            city=raw_address[address_column_mapping["city"]],
                            state=raw_address[address_column_mapping["state"]],
                            zipcode=raw_address[address_column_mapping["zipcode"]],
                        )
                        for raw_address in raw_addresses
                    ]
                    standard_addresses: List[
                        StandardizedAddress
                    ] = StandardizeAddr().standardize(
                        raw_addresses=raw_address_list,
                        cache_handler_obj=cache_handler,
                        vendor_obj=standardizing_vendor,
                    )
                    # map standard address back to a list of dictionary objects
                    standard_address_list: List[Dict[str, str]] = [
                        {
                            "address_id": standard_address.address.address_id,
                            f"{geolocation_column_prefix}latitude": standard_address.address.latitude,
                            f"{geolocation_column_prefix}longitude": standard_address.address.longitude,
                            address_column_mapping[
                                "line1"
                            ]: standard_address.address.line1,
                            address_column_mapping[
                                "line2"
                            ]: standard_address.address.line2,
                            address_column_mapping[
                                "city"
                            ]: standard_address.address.city,
                            address_column_mapping[
                                "state"
                            ]: standard_address.address.state,
                            address_column_mapping[
                                "zipcode"
                            ]: standard_address.address.zipcode,
                        }
                        for standard_address in standard_addresses
                    ]
                    return standard_address_list
                except Exception as standardize_exception:
                    print(f"standardize_exception: {standardize_exception}")
                    raise standardize_exception

            if not spark_is_data_frame_empty(address_df):
                # supply a schema whenever converting from rdd to df so spark does not have to infer the schema which
                # causes spark to inspect the data, which in this case causes standardize to be called multiple times
                result_df = address_df.rdd.mapPartitions(standardize).toDF(
                    schema=self.get_standardization_df_schema(geolocation_column_prefix)
                )
                combined_df = (
                    address_df.drop("raw_address")
                    .join(result_df, on="address_id")
                    .drop("address_id")
                )
                if func_get_response_path:
                    response_path: str = func_get_response_path(view)
                    self.logger.info(f"writing address data to {response_path}")
                    # kill the lineage by writing and reading back the data to avoid calling the standardize function more than once
                    combined_df.write.parquet(response_path)
                    standardized_df = df.sql_ctx.read.parquet(str(response_path))
                    standardized_df.createOrReplaceTempView(view)
                else:
                    # if no response path function is provided, just cache the results
                    combined_df.cache()

                    combined_df.createOrReplaceTempView(view)

        except Exception as e:
            self.logger.exception(str(e))
            raise
        return df

    def get_standardization_df_schema(
        self, geolocation_column_prefix: Optional[str]
    ) -> StructType:
        """
        Get the schema for the standardization dataframe with just the necessary geocoding columns and the join column.
        if the geolocation_column_prefix is provided, then the geolocation columns will be prefixed with the prefix.
        """
        latitude = f"{geolocation_column_prefix}latitude"
        longitude = f"{geolocation_column_prefix}longitude"

        return StructType(
            [
                StructField(latitude, StringType(), True),
                StructField(longitude, StringType(), True),
                StructField("address_id", StringType(), True),
            ]
        )

    # noinspection PyMethodMayBeStatic
    def get_address_columns_for_view(self, address_columns: List[str]) -> List[str]:
        mapping_list = ["address_id"]
        mapping_list.extend(address_columns)
        return mapping_list

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAddressColumnMapping(self) -> Dict[str, str]:
        return self.getOrDefault(self.address_column_mapping)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStandardizingVendor(self) -> StandardizingVendor:
        return self.getOrDefault(self.standardizing_vendor)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCacheHandler(self) -> CacheHandler:
        return self.getOrDefault(self.cache_handler)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getGeolocationColumnPrefix(self) -> Optional[str]:
        return self.getOrDefault(self.geolocation_column_prefix)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetResponsePath(self) -> Callable[[str], str] | None:
        return self.getOrDefault(self.func_get_response_path)



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType, StructField, StringType, DoubleType

from create_spark_session import clean_spark_session
from spark_pipeline_framework.transformers.framework_xml_loader.v1.framework_xml_loader import (
    FrameworkXmlLoader,
)


def test_can_load_xml_file(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test.xml')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkXmlLoader(
        view="my_view", file_path=test_file_path, row_tag="book"
    ).transform(df)

    result: DataFrame = spark_session.sql("SELECT * FROM my_view")
    result.show()
    assert result.count() == 12
    assert len(result.columns) == 7


def test_can_load_xml_file_with_schema(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test.xml')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    xml_shema = StructType(
        [
            StructField("_id", StringType(), True),
            StructField("author", StringType(), True),
            StructField("description", StringType(), True),
            StructField("genre", StringType(), True),
            StructField("price", DoubleType(), True),
            StructField("publish_date", StringType(), True),
            StructField("title", StringType(), True),
        ]
    )
    # Act
    FrameworkXmlLoader(
        view="my_view", file_path=test_file_path, row_tag="book", schema=xml_shema
    ).transform(df)

    result: DataFrame = spark_session.sql("SELECT * FROM my_view")
    result.show()
    assert result.count() == 12
    assert len(result.columns) == 7



from typing import List, Dict, Any

import pytest
from aioresponses import aioresponses

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)


@pytest.mark.asyncio
async def test_send_partition_request_to_server_async() -> None:
    parameters = get_fhir_receiver_parameters()

    rows: List[Dict[str, Any]] = [{"id": "1", "token": "abc"}]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for (
            result
        ) in FhirReceiverProcessor.send_partition_request_to_server_async(
            partition_index=0, rows=rows, parameters=parameters
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']



import json
from typing import Optional, Dict, Any, Union, List


class FhirReceiverException(Exception):
    def __init__(
        self,
        *,
        url: str,
        json_data: str,
        response_text: Optional[str],
        response_status_code: Optional[int],
        message: str,
        request_id: Optional[str],
    ) -> None:
        # Try to parse the json_data if possible
        parsed_json_data: Union[str, Dict[str, Any], List[Dict[str, Any]]] = json_data
        try:
            # Attempt to parse the json_data string into a dictionary
            parsed_json_data = json.loads(json_data)
        except (json.JSONDecodeError, TypeError):
            # If parsing fails, keep the original string
            pass

        # Prepare the error context dictionary
        error_context: Dict[str, Any] = {
            "message": f"FHIR receive failed: {message}",
            "url": url,
            "status_code": response_status_code,
            "response_text": response_text,
            # "json_data": parsed_json_data,
            "request_id": request_id,
        }

        # Store attributes for potential later access
        self.url: str = url
        self.data: Union[str, Dict[str, Any], List[Dict[str, Any]]] = parsed_json_data
        self.error_context: Dict[str, Any] = error_context

        # Custom string representation to improve readability
        error_message = self._format_error_message(error_context)

        # Call parent constructor with custom formatted message
        super().__init__(error_message)

    @staticmethod
    def _format_error_message(error_context: Dict[str, Any]) -> str:
        """
        Create a more readable error message with pretty-printed JSON if possible.

        Args:
            error_context (Dict[str, Any]): The error context dictionary

        Returns:
            str: Formatted error message
        """
        # Create a copy to avoid modifying the original context
        formatted_context = error_context.copy()

        # Try to pretty print JSON data if it's a dictionary
        # if isinstance(formatted_context['json_data'], dict) or isinstance(formatted_context['json_data'], list):
        #     try:
        #         # Use json.dumps with indentation for readability
        #         formatted_context['json_data'] = json.dumps(
        #             formatted_context['json_data'],
        #             indent=2,
        #             ensure_ascii=False
        #         )
        #     except Exception:
        #         # Fallback to str if JSON dumping fails
        #         formatted_context['json_data'] = str(formatted_context['json_data'])

        # Convert to a readable string format
        return "\n".join(
            [
                f"{key.replace('_', ' ').title()}\n{value}"
                for key, value in formatted_context.items()
                if value is not None
            ]
        )

    def __str__(self) -> str:
        """
        Override the default string representation to use our custom formatting.

        Returns:
            str: Formatted error message
        """
        return self._format_error_message(self.error_context)



from os import path, makedirs
from pathlib import Path
from shutil import rmtree

import pytest
from mockserver_client.mock_requests_loader import load_mock_source_api_json_responses
from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver import (
    FhirReceiver,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

from mockserver_client.mockserver_client import MockServerFriendlyClient


@pytest.mark.parametrize("run_synchronously", [True, False])
@pytest.mark.parametrize("use_data_streaming", [True, False])
def test_fhir_receiver_list(
    spark_session: SparkSession, run_synchronously: bool, use_data_streaming: bool
) -> None:
    # Arrange
    print()
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    patient_json_path: Path = temp_folder.joinpath("patient.json")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    test_name = "test_fhir_receiver_list"
    fhir_server_url = f"http://mock-server:1080/{test_name}/4_0_0"
    client = MockServerFriendlyClient("http://mock-server:1080")
    client.clear(test_name)

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("fhir_list_calls"),
        mock_client=client,
        url_prefix=f"{test_name}/4_0_0/Patient",
    )

    parameters = {"flow_name": "Test Pipeline V2", "team_name": "Data Operations"}

    # Act
    with ProgressLogger() as progress_logger:
        FhirReceiver(
            server_url=fhir_server_url,
            resource="Patient",
            file_path=patient_json_path,
            progress_logger=progress_logger,
            parameters=parameters,
            run_synchronously=run_synchronously,
            use_data_streaming=use_data_streaming,
        ).transform(df)

    # Assert
    json_df: DataFrame = df.sparkSession.read.json(str(patient_json_path))
    json_df.show()
    json_df.printSchema()

    assert json_df.select("resourceType").collect()[0][0] == "Patient"



from __future__ import annotations

import asyncio
from abc import abstractmethod
from datetime import datetime
from logging import Logger
from typing import (
    Iterator,
    AsyncIterator,
    List,
    Optional,
    cast,
    AsyncGenerator,
    Any,
    Dict,
)

# noinspection PyPackageRequirements
import pandas as pd
from helix_fhir_client_sdk.utilities.async_parallel_processor.v1.async_parallel_processor import (
    AsyncParallelProcessor,
    ParallelFunctionContext,
)

# noinspection PyPackageRequirements
from pyspark import TaskContext

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.chunk_container import (
    ChunkContainer,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.event_handlers import (
    OnPartitionStartEventHandler,
    OnChunkStartEventHandler,
    OnChunkCompletionEventHandler,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.function_types import (
    HandlePandasBatchFunction,
    AcceptedParametersType,
    AcceptedColumnDataType,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.partition_context import (
    PartitionContext,
)

AcceptedDataSourceType = pd.DataFrame | pd.Series  # type:ignore[type-arg]


class AsyncBasePandasUDF[
    TParameters: AcceptedParametersType,
    TInputDataSource: AcceptedDataSourceType,
    TOutputDataSource: AcceptedDataSourceType,
    TInputColumnDataType: AcceptedColumnDataType,
    TOutputColumnDataType: AcceptedColumnDataType,
]:
    """
    This base class implements the logic to run an async function in Spark using Pandas UDFs.


    TInputDataSource is the type of the input data.  It can be a pd.DataFrame or pd.Series.
    TOutputDataSource is the type of the output data. It can be a pd.DataFrame or pd.Series.
    TInputColumnDataType is the type of the data in the incoming dictionaries.  It can be dict for struct columns
        otherwise the type of the column
    TOutputColumnDataType is the type of the data in the outgoing dictionaries. It can be dict for struct columns
        otherwise the type of the column
    """

    def __init__(
        self,
        *,
        async_func: HandlePandasBatchFunction[
            TParameters, TInputColumnDataType, TOutputColumnDataType
        ],
        parameters: Optional[TParameters],
        pandas_udf_parameters: AsyncPandasUdfParameters,
    ) -> None:
        """
        This class wraps an async function in a Pandas UDF for use in Spark.  The subclass must
        supply TInputDataSource and implement the abstract methods in this class

        :param async_func: an async function that takes a list of dictionaries as input and
                            returns a list of dictionaries
        :param parameters: parameters to pass to the async function
        :param pandas_udf_parameters: the parameters for the Pandas UDF
        """
        self.async_func: HandlePandasBatchFunction[
            TParameters, TInputColumnDataType, TOutputColumnDataType
        ] = async_func
        self.parameters: Optional[TParameters] = parameters
        self.pandas_udf_parameters: AsyncPandasUdfParameters = pandas_udf_parameters
        self.log_level = (
            pandas_udf_parameters.log_level
            if pandas_udf_parameters.log_level
            else "INFO"
        )
        self.logger: Logger = get_logger(__name__, self.log_level)
        self._on_partition_start_event_handler: OnPartitionStartEventHandler | None = (
            None
        )
        self._on_partition_completion_event_handler: (
            OnPartitionStartEventHandler | None
        ) = None
        self._on_chunk_start_event_handler: OnChunkStartEventHandler | None = None
        self._on_chunk_completion_event_handler: (
            OnChunkCompletionEventHandler | None
        ) = None

    @property
    def on_partition_start(self) -> OnPartitionStartEventHandler | None:
        return self._on_partition_start_event_handler

    @on_partition_start.setter
    def on_partition_start(self, value: OnPartitionStartEventHandler | None) -> None:
        self._on_partition_start_event_handler = value

    @property
    def on_partition_completion(self) -> OnPartitionStartEventHandler | None:
        return self._on_partition_completion_event_handler

    @on_partition_completion.setter
    def on_partition_completion(
        self, value: OnPartitionStartEventHandler | None
    ) -> None:
        self._on_partition_completion_event_handler = value

    @property
    def on_chunk_start(self) -> OnChunkStartEventHandler | None:
        return self._on_chunk_start_event_handler

    @on_chunk_start.setter
    def on_chunk_start(self, value: OnChunkStartEventHandler | None) -> None:
        self._on_chunk_start_event_handler = value

    @property
    def on_chunk_completion(self) -> OnChunkCompletionEventHandler | None:
        return self._on_chunk_completion_event_handler

    @on_chunk_completion.setter
    def on_chunk_completion(self, value: OnChunkCompletionEventHandler | None) -> None:
        self._on_chunk_completion_event_handler = value

    @staticmethod
    async def to_async_iter(
        sync_iter: Iterator[TInputDataSource],
    ) -> AsyncIterator[TInputDataSource]:
        item: TInputDataSource
        for item in sync_iter:
            yield item

    async def get_chunks_of_size(
        self, *, chunk_size: int, chunk_iterator: AsyncIterator[TInputDataSource]
    ) -> AsyncGenerator[List[TInputColumnDataType], None]:
        """
        Given an async iterator of dataframes, this function will yield chunks of the content of the dataframes.

        :param chunk_size: the size of the chunks
        :param chunk_iterator: the async iterator of chunks
        :return: an async generator of chunks of dictionaries
        """
        chunk_number: int = 0
        chunk_input_values: List[TInputColumnDataType] = []

        chunk: TInputDataSource
        async for chunk in chunk_iterator:
            chunk_number += 1
            # Convert JSON strings to dictionaries
            # convert the dataframe to a list of dictionaries
            input_values = await self.get_input_values_from_chunk(chunk)
            chunk_input_values.extend(input_values)
            if len(chunk_input_values) >= chunk_size:
                # print(f"yielding chunk_input_values at chunk {chunk_number}: {len(chunk_input_values)}"
                #       f", chunk_size: {chunk_size}")
                current_chunk_values: List[TInputColumnDataType] = chunk_input_values[
                    :chunk_size
                ]
                # remove chunk from chunk_input_values
                chunk_input_values = chunk_input_values[chunk_size:]
                yield current_chunk_values
        if len(chunk_input_values) > 0:
            # print(f"yielding chunk_input_values at the end at chunk {chunk_number}: {len(chunk_input_values)}"
            #       f", chunk_size: {chunk_size}")
            yield chunk_input_values

    @abstractmethod
    async def get_input_values_from_chunk(
        self, chunk: TInputDataSource
    ) -> List[TInputColumnDataType]:
        """
        This abstract method is called to convert the input data to a list of dictionaries.

        :param chunk: the input data
        :return: a list of dictionaries
        """
        ...

    @staticmethod
    async def get_chunk_containers(
        chunks: AsyncGenerator[List[TInputColumnDataType], None],
    ) -> AsyncGenerator[ChunkContainer[TInputColumnDataType], None]:
        """
        create a list of chunk containers where each container contains the chunk and metadata
        like chunk index and chunk input values index


        :param chunks: async generator of chunks
        :return: async generator of chunk containers
        """
        chunk_index: int = 0
        chunk_input_values_index: int = 0
        async for chunk_input_values in chunks:
            chunk_index += 1
            begin_chunk_input_values_index: int = chunk_input_values_index + 1
            chunk_input_values_index += len(chunk_input_values)
            chunk_container: ChunkContainer[TInputColumnDataType] = ChunkContainer(
                chunk_input_values=chunk_input_values,
                chunk_index=chunk_index,
                begin_chunk_input_values_index=begin_chunk_input_values_index,
                end_chunk_input_values_index=chunk_input_values_index,
            )
            yield chunk_container

    async def process_partition_async(
        self,
        chunk_iterator: Iterator[TInputDataSource],
        async_parallel_processor: Optional[AsyncParallelProcessor] = None,
        **kwargs: Any,
    ) -> AsyncIterator[TOutputDataSource]:
        """
        Apply the custom function `self.async_func` to the input chunk iterator asynchronously.
        This is an async function that processes the input data in chunks.
        Override if you want to do more when processing a partition.  Use kwargs to pass additional parameters.
        These parameters will be passed to the async function.


        :param chunk_iterator: iterator of chunks of input data
        :param kwargs: additional parameters
        :param async_parallel_processor: The async parallel processor to use.  If not set, a new one will be created.
                                        You may want to pass one in if you want to share the task pool with other work so
                                        max_concurrent_tasks is respected across all work.
        :return: async iterator of chunks of output data
        """

        if async_parallel_processor is None:
            async_parallel_processor = AsyncParallelProcessor(
                name="process_partition_async",
                max_concurrent_tasks=self.pandas_udf_parameters.maximum_concurrent_tasks,
            )
        task_context: Optional[TaskContext] = TaskContext.get()
        partition_index: int = task_context.partitionId() if task_context else 0
        partition_start_time: datetime = datetime.now()

        if self._on_partition_start_event_handler is not None:
            await self._on_partition_start_event_handler(
                partition_index=partition_index
            )

        chunk_input_values: List[TInputColumnDataType]
        chunks: AsyncGenerator[List[TInputColumnDataType], None] = (
            self.get_chunks_of_size(
                chunk_size=self.pandas_udf_parameters.max_chunk_size,
                chunk_iterator=self.to_async_iter(chunk_iterator),
            )
        )

        if self.pandas_udf_parameters.process_chunks_in_parallel:
            self.logger.debug(
                f"Starting process_partition_async | partition: {partition_index} | type: parallel"
            )
            async for result in self.process_chunks_async_in_parallel(
                partition_context=PartitionContext(
                    partition_index=partition_index,
                    partition_start_time=partition_start_time,
                ),
                chunk_containers=AsyncBasePandasUDF.get_chunk_containers(chunks),
                async_parallel_processor=async_parallel_processor,
                **kwargs,
            ):
                yield result
        else:
            self.logger.debug(
                f"Starting process_partition_async | partition: {partition_index} | type: sequential"
            )
            async for result in self.process_chunks_sequential_async(
                partition_context=PartitionContext(
                    partition_index=partition_index,
                    partition_start_time=partition_start_time,
                ),
                chunk_containers=AsyncBasePandasUDF.get_chunk_containers(chunks),
                **kwargs,
            ):
                yield result

        self.logger.debug(
            f"Finished process_partition_async | partition {partition_index}"
        )
        if self._on_partition_completion_event_handler is not None:
            await self._on_partition_completion_event_handler(
                partition_index=partition_index
            )

    async def process_chunks_sequential_async(
        self,
        *,
        partition_context: PartitionContext,
        chunk_containers: AsyncGenerator[ChunkContainer[TInputColumnDataType], None],
        **kwargs: Any,
    ) -> AsyncGenerator[TOutputDataSource, None]:
        """
        This function processes the chunks of input data asynchronously.  You can override this function
        to add additional processing logic.


        :param partition_context: the context for the partition
        :param chunk_containers: the async generator of chunk containers
        :param kwargs: additional parameters
        :return: an async generator of output data
        """

        chunk_container: ChunkContainer[TInputColumnDataType]
        async for chunk_container in chunk_containers:
            output_data: TOutputDataSource = await self.process_chunk_container(
                chunk_container=chunk_container,
                partition_context=partition_context,
                **kwargs,
            )
            yield output_data

    async def process_chunk_container(
        self,
        *,
        chunk_container: ChunkContainer[TInputColumnDataType],
        partition_context: PartitionContext,
        **kwargs: Any,
    ) -> TOutputDataSource:
        """
        This function processes a single chunk of input data asynchronously.
        Override this to pass additional parameters to the async function.


        :param chunk_container: the chunk container
        :param partition_context: the context for the partition
        :param kwargs: additional parameters
        :return: the output data
        """

        chunk_start_time: datetime = datetime.now()
        if self._on_chunk_start_event_handler is not None:
            await self._on_chunk_start_event_handler(
                partition_index=partition_context.partition_index,
                chunk_index=chunk_container.chunk_index,
                chunk_range=range(
                    chunk_container.begin_chunk_input_values_index,
                    chunk_container.end_chunk_input_values_index,
                ),
                chunk_start_time=chunk_start_time,
            )
        self.logger.debug(
            f"Starting process_chunk_container | partition: {partition_context.partition_index}"
            f" | Chunk: {chunk_container.chunk_index}"
            f" | range: {chunk_container.begin_chunk_input_values_index}-{chunk_container.end_chunk_input_values_index}"
        )
        if len(chunk_container.chunk_input_values) == 0:
            self.logger.debug(
                f"Empty chunk | partition: {partition_context.partition_index}"
                f" | Chunk: {chunk_container.chunk_index}"
            )
            return await self.create_output_from_dict([])
        else:
            output_values: List[TOutputColumnDataType] = []
            chunk_input_range: range = range(
                chunk_container.begin_chunk_input_values_index,
                chunk_container.end_chunk_input_values_index,
            )
            async for output_value in cast(
                AsyncGenerator[
                    TOutputColumnDataType, None
                ],  # mypy can't figure out that self.async_func is an async generator
                self.async_func(
                    run_context=AsyncPandasBatchFunctionRunContext(
                        partition_index=partition_context.partition_index,
                        chunk_index=chunk_container.chunk_index,
                        chunk_input_range=chunk_input_range,
                        partition_start_time=partition_context.partition_start_time,
                    ),
                    input_values=chunk_container.chunk_input_values,
                    parameters=self.parameters,
                    additional_parameters=kwargs,
                ),
            ):
                output_values.append(output_value)
            self.logger.debug(
                f"Finished process_chunk_container | partition: {partition_context.partition_index}"
                f" | Chunk: {chunk_container.chunk_index}"
                f" | range: {chunk_container.begin_chunk_input_values_index}-{chunk_container.end_chunk_input_values_index}"
            )
            chunk_end_time: datetime = datetime.now()
            chunk_duration: float = (chunk_end_time - chunk_start_time).total_seconds()
            if self._on_chunk_completion_event_handler is not None:
                await self._on_chunk_completion_event_handler(
                    partition_index=partition_context.partition_index,
                    chunk_index=chunk_container.chunk_index,
                    chunk_range=range(
                        chunk_container.begin_chunk_input_values_index,
                        chunk_container.end_chunk_input_values_index,
                    ),
                    chunk_start_time=chunk_start_time,
                    chunk_end_time=chunk_end_time,
                    chunk_duration=chunk_duration,
                )
            return await self.create_output_from_dict(output_values)

    async def process_chunks_async_in_parallel(
        self,
        *,
        partition_context: PartitionContext,
        chunk_containers: AsyncGenerator[ChunkContainer[TInputColumnDataType], None],
        async_parallel_processor: AsyncParallelProcessor,
        **kwargs: Any,
    ) -> AsyncGenerator[TOutputDataSource, None]:
        """
        This function processes the chunks of input data asynchronously in parallel.  You can override this function
        to add additional processing logic.


        :param partition_context: the context for the partition
        :param chunk_containers: the async generator of chunk containers
        :param kwargs: additional parameters
        :param async_parallel_processor: The async parallel processor to use.
        :return: an async generator of output data
        """

        # read the whole list from async generator so we can kick off all the tasks in parallel
        chunk_containers_list: List[ChunkContainer[TInputColumnDataType]] = [
            x async for x in chunk_containers
        ]

        if len(chunk_containers_list) == 0:
            self.logger.debug(
                f"Empty chunk list | partition: {partition_context.partition_index}"
            )
            return

        self.logger.debug(
            f"Starting process_chunks_async_in_parallel | partition: {partition_context.partition_index}"
            f" | chunk_count: {len(chunk_containers_list)}"
            f" | range: {chunk_containers_list[0].begin_chunk_input_values_index}-{chunk_containers_list[-1].end_chunk_input_values_index}"
        )

        # noinspection PyShadowingNames,PyUnusedLocal
        async def process_chunk_container_fn(
            context: ParallelFunctionContext,
            row: ChunkContainer[TInputColumnDataType],
            parameters: Optional[TParameters],
            additional_parameters: Optional[Dict[str, Any]],
        ) -> TOutputDataSource:
            """This functions wraps the process_chunk_container function which is an instance method"""

            self.logger.debug(
                f"Starting process_chunk_container_fn | partition: {partition_context.partition_index}"
                f" | Chunk: {row.chunk_index}"
                f" | range: {row.begin_chunk_input_values_index}-{row.end_chunk_input_values_index}"
                f" | task_index: {context.task_index}"
                f" | total_task_count: {context.total_task_count}"
                f" | parallel_processor: {context.name}"
            )
            result: TOutputDataSource = await self.process_chunk_container(
                chunk_container=row, partition_context=partition_context, **kwargs
            )
            self.logger.debug(
                f"Finished process_chunk_container_fn | partition: {partition_context.partition_index}"
                f" | Chunk: {row.chunk_index}"
                f" | range: {row.begin_chunk_input_values_index}-{row.end_chunk_input_values_index}"
                f" | task_index: {context.task_index}"
                f" | total_task_count: {context.total_task_count}"
                f" | parallel_processor: {context.name}"
            )
            return result

        # now run all the tasks in parallel yielding the results as they get ready
        result: TOutputDataSource
        parameters: TParameters | None = self.parameters

        assert async_parallel_processor is not None
        async for result in async_parallel_processor.process_rows_in_parallel(
            rows=chunk_containers_list,
            process_row_fn=process_chunk_container_fn,
            max_concurrent_tasks=self.pandas_udf_parameters.maximum_concurrent_tasks,
            parameters=parameters,
            log_level=self.pandas_udf_parameters.log_level,
            **kwargs,
        ):
            yield result

        self.logger.debug(
            f"Finished process_chunks_async_in_parallel |  partition: {partition_context.partition_index}"
            f" | chunk_count: {len(chunk_containers_list)}"
            f" | range: {chunk_containers_list[0].begin_chunk_input_values_index}-{chunk_containers_list[-1].end_chunk_input_values_index}"
        )

    @abstractmethod
    async def create_output_from_dict(
        self, output_values: List[TOutputColumnDataType]
    ) -> TOutputDataSource:
        """
        This abstract method is called to convert the output data from a list of dictionaries to the output data type.

        :param output_values: the output data
        :return: the output data
        """
        ...

    # noinspection PyMethodMayBeStatic
    async def collect_async_iterator(
        self, async_iter: AsyncIterator[TOutputDataSource]
    ) -> List[TOutputDataSource]:
        """
        This function collects the output of an async iterator into a list.

        :param async_iter: the async iterator
        :return: a list of the output
        """
        return [item async for item in async_iter]

    def apply_process_partition_udf(
        self, chunk_iter: Iterator[TInputDataSource]
    ) -> Iterator[TOutputDataSource]:
        """
        This function will be called for each partition in Spark.  It will run on worker nodes in parallel.
        Within each partition, the input data will be processed in chunks using Pandas.  The size of the chunks
        is controlled by the `spark.sql.execution.arrow.maxRecordsPerBatch` configuration.

        :param chunk_iter: Iterable[pd.DataFrame | pd.Series]
        :return: Iterable[pd.DataFrame | pd.Series]
        """
        self.logger.setLevel(self.log_level)
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        async_iter: AsyncIterator[TOutputDataSource] = self.process_partition_async(
            chunk_iter, loop=loop
        )
        async_gen = loop.run_until_complete(self.collect_async_iterator(async_iter))
        return iter(async_gen)



import os
from typing import Any, Dict, List, Optional, Callable

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import (
    to_json,
    struct,
    when,
    col,
    lit,
)
from pyspark.sql.types import StructType, StructField, StringType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.address_standardization.v2.address_standardization_processor import (
    AddressStandardizationParameters,
    AddressStandardizationProcessor,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_struct_column_to_struct_udf import (
    AsyncPandasStructColumnToStructColumnUDF,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class AddressStandardization(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        address_column_mapping: Dict[str, str],
        standardizing_vendor: StandardizingVendor[BaseVendorApiResponse],
        cache_handler: CacheHandler,
        func_get_response_path: Callable[[str], str] | None = None,
        geolocation_column_prefix: Optional[str] = "",
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        batch_size: int = 100,
    ):
        """
        Standardize and geocode addresses in a view using the specified standardizing vendor. Address columns in the view
        should be mapped to the standard address columns like the following example. the values in the dictionary should
        be the column names in the view.
        {
            "line1": "address1",
            "line2": "address2",
            "city": "city",
            "state": "state",
            "zipcode": "zip"
        }

        :param view: view with the address data to standardize
        :param address_column_mapping: mapping of address columns to standardize
        :param standardizing_vendor: vendor class to use for standardization
        :param cache_handler: cache handler to use
        :param func_get_response_path: function to get response path, this should be used if the data in the view is large,
        if it is not specified then the results will be cached
        :param geolocation_column_prefix: prefix for geolocation columns
        :param name: name for the stage in spark
        :param parameters: additional parameters
        :param progress_logger: progress logger
        :param batch_size: batch size for standardization
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.address_column_mapping: Param[Dict[str, str]] = Param(
            self, "address_column_mapping", ""
        )
        self._setDefault(address_column_mapping=address_column_mapping)

        self.standardizing_vendor: Param[StandardizingVendor[BaseVendorApiResponse]] = (
            Param(self, "standardizing_vendor", "")
        )
        self._setDefault(view=standardizing_vendor)

        self.cache_handler: Param[CacheHandler] = Param(self, "cache_handler", "")
        self._setDefault(view=cache_handler)

        self.geolocation_column_prefix: Param[Optional[str]] = Param(
            self, "geolocation_column_prefix", ""
        )
        self._setDefault(geolocation_column_prefix=geolocation_column_prefix)

        self.func_get_response_path: Param[Callable[[str], str]] = Param(
            self, "func_get_response_path", ""
        )
        self._setDefault(func_get_response_path=func_get_response_path)

        self.batch_size: Param[int] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        batch_size: int = self.getOrDefault(self.batch_size)
        try:
            view: str = self.getView()
            # where to write the results
            func_get_response_path: Callable[[str], str] | None = (
                self.getFuncGetResponsePath()
            )
            # input columns
            address_column_mapping: Dict[str, str] = self.getAddressColumnMapping()
            # list of column names
            address_column_mapping_names: List[str] = list(
                address_column_mapping.values()
            )
            # what vendor to use for standardizing
            standardizing_vendor: StandardizingVendor[BaseVendorApiResponse] = (
                self.getStandardizingVendor()
            )
            # what cache handler to use
            cache_handler: CacheHandler = self.getCacheHandler()
            # what column prefix to use for lat/long columns (if any)
            geolocation_column_prefix: Optional[str] = self.getGeolocationColumnPrefix()

            # Load view into a dataframe
            address_df: DataFrame = df.sql_ctx.table(view)

            # see if we need to partition the incoming dataframe
            total_partitions: int = address_df.rdd.getNumPartitions()

            # create a `raw_address` column to store all the address parts as a json string
            address_df = address_df.withColumn(
                # create a new column that contains all the address parts as json string
                "raw_address",
                to_json(
                    struct(
                        *[
                            when(col(x).isNotNull(), col(x)).otherwise(lit("")).alias(x)
                            for x in address_column_mapping_names
                        ]
                    )
                ),
            )
            incoming_row_count: int = address_df.count()

            # if there are rows then standardize them
            if not spark_is_data_frame_empty(address_df):
                # add a new column that will hold the result of standardization as a struct
                combined_df = address_df.withColumn(
                    colName="standardized_address",
                    col=AsyncPandasStructColumnToStructColumnUDF(
                        async_func=AddressStandardizationProcessor.standardize_list,  # type: ignore[arg-type]
                        parameters=AddressStandardizationParameters(
                            total_partitions=total_partitions,
                            log_level=os.getenv("LOGLEVEL", "INFO"),
                            address_column_mapping=address_column_mapping,
                            standardizing_vendor=standardizing_vendor,
                            cache_handler=cache_handler,
                            geolocation_column_prefix=geolocation_column_prefix,
                        ),
                        pandas_udf_parameters=AsyncPandasUdfParameters(
                            max_chunk_size=batch_size
                        ),
                    ).get_pandas_udf(
                        return_type=self.get_standardization_df_schema(
                            address_column_mapping=address_column_mapping,
                            geolocation_column_prefix=geolocation_column_prefix,
                        )
                    )(
                        address_df["raw_address"]
                    ),
                )
                # extract the latitude and longitude from the struct and add them as separate columns
                combined_df = combined_df.select(
                    col("*"),
                    col(f"standardized_address.{geolocation_column_prefix}latitude"),
                    col(f"standardized_address.{geolocation_column_prefix}longitude"),
                ).drop("standardized_address")

                assert (
                    incoming_row_count == combined_df.count()
                ), "Address Standardization should not increase the row count"

                if func_get_response_path:
                    response_path: str = func_get_response_path(view)
                    self.logger.info(f"writing address data to {response_path}")
                    # kill the lineage by writing and reading back the data to avoid
                    # calling the standardize function more than once
                    combined_df.write.parquet(response_path)
                    standardized_df = df.sql_ctx.read.parquet(str(response_path))
                    standardized_df.createOrReplaceTempView(view)
                else:
                    # if no response path function is provided, just cache the results
                    combined_df.cache()

                    combined_df.createOrReplaceTempView(view)

        except Exception as e:
            self.logger.exception(str(e))
            raise
        return df

    @staticmethod
    def get_standardization_df_schema(
        *,
        address_column_mapping: Dict[str, str],
        geolocation_column_prefix: Optional[str],
    ) -> StructType:
        """
        Get the schema for the standardization dataframe with just the necessary geocoding columns and the join column.
        if the geolocation_column_prefix is provided, then the geolocation columns will be prefixed with the prefix.
        """
        latitude = f"{geolocation_column_prefix}latitude"
        longitude = f"{geolocation_column_prefix}longitude"

        return StructType(
            [
                StructField(latitude, StringType(), True),
                StructField(longitude, StringType(), True),
                StructField(address_column_mapping["line1"], StringType(), True),
                StructField(address_column_mapping["line2"], StringType(), True),
                StructField(address_column_mapping["city"], StringType(), True),
                StructField(address_column_mapping["state"], StringType(), True),
                StructField(address_column_mapping["zipcode"], StringType(), True),
            ]
        )

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAddressColumnMapping(self) -> Dict[str, str]:
        return self.getOrDefault(self.address_column_mapping)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStandardizingVendor(self) -> StandardizingVendor[BaseVendorApiResponse]:
        return self.getOrDefault(self.standardizing_vendor)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getCacheHandler(self) -> CacheHandler:
        return self.getOrDefault(self.cache_handler)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getGeolocationColumnPrefix(self) -> Optional[str]:
        return self.getOrDefault(self.geolocation_column_prefix)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetResponsePath(self) -> Callable[[str], str] | None:
        return self.getOrDefault(self.func_get_response_path)



# mypy: ignore-errors

import json
from datetime import datetime
from typing import List, Dict, cast

import pymongo
from pymongo import MongoClient
from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.connecthub_data_receiver.v1.connecthub_data_receiver import (
    ConnectHubDataReceiver,
)

from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_connecthub_data_receiver(spark_session: SparkSession) -> None:
    # Arrange
    test_name = "connecthub_data_receiver_test"
    page_size = 0
    last_run_date = datetime.strptime("1970-01-01", "%Y-%m-%d")
    conn_string = "mongodb://mongo:27017/"
    db_name = "integration_hub"

    mongo_client: MongoClient = pymongo.MongoClient(conn_string)
    integration_hub = mongo_client[db_name]
    client_connection = integration_hub.client_connection

    # delete the collection if it's already there
    if client_connection is not None:
        client_connection.drop()

    client_connection.insert_one(
        {
            "displayLabel": "Should Return",
            "lastUpdatedOnDate": datetime.now(),
            "isRestricted": False,
        }
    )
    client_connection.insert_one(
        {
            "displayLabel": "Should not Return Restricted",
            "lastUpdatedOnDate": datetime.now(),
            "isRestricted": True,
        }
    )

    client_connection.insert_one(
        {
            "displayLabel": "Should NOT Return",
            "lastUpdatedOnDate": datetime.strptime("1900-01-01", "%Y-%m-%d"),
            "isRestricted": False,
        }
    )

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # Act
    with ProgressLogger() as progress_logger:
        ConnectHubDataReceiver(
            view_name=test_name,
            db_name=db_name,
            progress_logger=progress_logger,
            page_size=page_size,
            conn_string=conn_string,
            last_run_date=last_run_date,
            query_parameters={"isRestricted": False},
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table(test_name)
    result_df.printSchema()
    result_df.show(truncate=False)
    result = (
        result_df.toJSON()
        .map(lambda j: cast(Dict[str, Dict[str, List[str]]], json.loads(j)))
        .collect()
    )
    assert len(result) == 1
    assert result[0].get("displayLabel") is not "Should NOT Return"
    assert result[0].get("displayLabel") is not "Should not Return Restricted"



import pymongo
from pymongo import MongoClient, UpdateOne
from pymongo.collection import Collection
from typing import List, Dict, Any, Set, Tuple, Generator, Sequence, Iterable
import pathlib
import glob
import json
import csv

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response import (
    VendorResponse,
)

# define column mappings
column_mappings = {
    "address_catalog": {
        "line1": "line1",
        "line2": "line2",
        "city": "city",
        "state": "state",
        "zipcode": "zipcode",
    },
    "nppes_npi_pfile_mailing": {
        "line1": "Provider First Line Business Mailing Address",
        "line2": "Provider Second Line Business Mailing Address",
        "city": "Provider Business Mailing Address City Name",
        "state": "Provider Business Mailing Address State Name",
        "zipcode": "Provider Business Mailing Address Postal Code",
    },
    "nppes_npi_pfile_practice": {
        "line1": "Provider First Line Business Practice Location Address",
        "line2": "Provider Second Line Business Practice Location Address",
        "city": "Provider Business Practice Location Address City Name",
        "state": "Provider Business Practice Location Address State Name",
        "zipcode": "Provider Business Practice Location Address Postal Code",
    },
    "nppes_practice_location_pfile": {
        "line1": "Provider Secondary Practice Location Address- Address Line 1",
        "line2": "Provider Secondary Practice Location Address-  Address Line 2",
        "city": "Provider Secondary Practice Location Address - City Name",
        "state": "Provider Secondary Practice Location Address - State Name",
        "zipcode": "Provider Secondary Practice Location Address - Postal Code",
    },
}

# set base directory path
base_directory = pathlib.Path(__file__).parent.joinpath("./")
input_directory = base_directory / "input"
output_directory = base_directory / "output"


# adds geocoded geocodio metadata to file based on passed in address column mapping
def add_geocodio_metadata(
    input_file_path: pathlib.Path,
    geocodio_catalog_file: pathlib.Path,
    address_columns: Dict[str, str],
) -> None:
    # store unique addresses in dict with indexed key for easy lookup
    address_dict = {}
    with open(geocodio_catalog_file, mode="r", newline="") as geocodio_catalog:
        # expected header:
        # [
        #   geocodio_line1,geocodio_city,geocodio_state,geocodio_postal_code,Latitude,Longitude,
        #   Accuracy Score,Accuracy Type,Number,Street,Unit Type,Unit Number,City,State,County,Zip,Country,Source
        # ]
        geocodio_catalog_reader = csv.DictReader(geocodio_catalog)
        for row in geocodio_catalog_reader:
            key = (
                # for some reason the linter doesn't like strings here so type ignore
                row["geocodio_line1"],
                row["geocodio_city"],
                row["geocodio_state"],
                row["geocodio_postal_code"][:5],
            )
            address_dict[key] = row

    input_file_name = input_file_path.stem

    if geocodio_catalog_reader.fieldnames is None:
        print("Geocodio Catalog file is missing fieldnames")
        return

    geocodio_meta_fieldnames = [
        field.replace(" ", "_")
        for field in geocodio_catalog_reader.fieldnames
        if "geocodio" not in field
    ]

    # open a writer for addresses in the file that are missing a geocodio entry
    missing_addresses_path = output_directory / "missing_addresses"
    missing_addresses_file = (
        missing_addresses_path / f"{input_file_name}_missing_addresses.csv"
    )
    missing_addresses_file.parent.mkdir(parents=True, exist_ok=True)
    with missing_addresses_file.open(mode="w", newline="") as out_missing:
        missing_writer = csv.DictWriter(
            out_missing,
            fieldnames=["addr_line_1", "city", "state", "postal_code"],
            quoting=csv.QUOTE_ALL,
        )
        missing_writer.writeheader()
        missing = set()

        output_file_path = output_directory / "geocoded_geocodio_files"
        output_file = output_file_path / f"{input_file_name}_geocoded.csv"
        output_file.parent.mkdir(parents=True, exist_ok=True)

        # open both the input and output files
        with (
            input_file_path.open(mode="r", newline="") as input_file,
            output_file.open(mode="w", newline="") as out_file,
        ):

            input_reader = csv.DictReader(input_file)

            if input_reader.fieldnames is None:
                print("Input file is missing fieldnames")
                return

            input_fieldnames_without_lat_long = [
                field
                for field in input_reader.fieldnames
                if field not in ["Latitude", "Longitude"]
            ]

            writer_fieldnames = []
            for field in geocodio_meta_fieldnames:
                if field in ["Latitude", "Longitude"]:
                    writer_fieldnames.append(field)
                else:
                    writer_fieldnames.append(f"geocodio_{field}")

            writer = csv.DictWriter(
                out_file,
                fieldnames=input_fieldnames_without_lat_long + writer_fieldnames,
                quoting=csv.QUOTE_ALL,
            )

            writer.writeheader()
            for row in input_reader:
                # pull key from input file
                key = (
                    row[address_columns["line1"]],
                    row[address_columns["city"]],
                    row[address_columns["state"]],
                    row[address_columns["zipcode"]][:5],
                )
                # write row with all geocodio fields from lookup dict
                if key in address_dict.keys():
                    row_copy = dict(row)

                    # just in case delete lat/long from input file if they're there
                    if "Latitude" in row_copy.keys():
                        del row_copy["Latitude"]
                    if "Longitude" in row_copy.keys():
                        del row_copy["Longitude"]

                    for field in geocodio_meta_fieldnames:
                        if field in ["Latitude", "Longitude"]:
                            row_copy[field] = address_dict[key][field]
                        else:
                            row_copy[f"geocodio_{field}"] = address_dict[key][
                                field.replace("_", " ")
                            ]

                    writer.writerow(row_copy)

                # write missing address if not found
                else:
                    if key not in missing:
                        missing.add(key)
                        missing_writer.writerow(
                            {
                                "addr_line_1": key[0],
                                "city": key[1],
                                "state": key[2],
                                "postal_code": key[3],
                            }
                        )


# finds cases where the same address hash in a file has different lat/longs
def save_address_collisions(
    glob_string: str, address_column_mapping: Dict[str, str]
) -> None:
    address_collisions: Dict[str, Dict[str, Dict[str, List[int]]]] = {}
    file_type = glob_string.split("/")[-2]

    # check all files in a directory
    for csv_filename in glob.glob(glob_string, recursive=True):
        with open(csv_filename, mode="r") as csv_file:
            csv_reader = csv.DictReader(csv_file)
            count = 1

            for row in csv_reader:
                # store address column keys
                address_key = (
                    row[address_column_mapping["line1"]],
                    row[address_column_mapping["line2"]],
                    row[address_column_mapping["city"]],
                    row[address_column_mapping["state"]],
                    row[address_column_mapping["zipcode"]],
                )

                # store lat and long
                lat_long_key = (
                    row["Latitude"],
                    row["Longitude"],
                )

                # if the keys for different lat/longs are found save them again
                if str(address_key) in address_collisions.keys():
                    if str(lat_long_key) in address_collisions[str(address_key)].keys():
                        if (
                            csv_filename
                            in address_collisions[str(address_key)][
                                str(lat_long_key)
                            ].keys()
                        ):
                            address_collisions[str(address_key)][str(lat_long_key)][
                                csv_filename
                            ].append(count)
                        else:
                            address_collisions[str(address_key)][str(lat_long_key)][
                                csv_filename
                            ] = [count]

                    else:
                        address_collisions[str(address_key)][str(lat_long_key)] = {
                            csv_filename: [count]
                        }

                # if we haven't added these keys yet, do that here
                else:
                    address_collisions[str(address_key)] = {
                        str(lat_long_key): {csv_filename: [count]}
                    }

                count += 1

    # write output to file
    output_file = (
        output_directory / "address_collisions" / f"{file_type}_address_collisions.json"
    )
    output_file.parent.mkdir(parents=True, exist_ok=True)
    with output_file.open(mode="w") as outfile:
        json.dump(address_collisions, outfile, indent=2)


# pass in list of columns to write new file with only those columns
def extract_csv_columns(
    input_file_path: str, output_file_path: str, columns_to_extract: List[str]
) -> None:
    with (
        open(input_file_path, mode="r") as infile,
        open(output_file_path, mode="w", newline="") as outfile,
    ):
        reader = csv.DictReader(infile)
        writer = csv.DictWriter(outfile, fieldnames=columns_to_extract)
        writer.writeheader()
        for row in reader:
            writer.writerow({column: row[column] for column in columns_to_extract})


# pass in a reader object and generate iterations based on chunk_size
def batch_process_csv_reader(
    reader: Iterable[Any], chunk_size: int
) -> Generator[Any, None, None]:
    chunk = []
    for row in reader:
        chunk.append(row)
        if len(chunk) == chunk_size:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


# pulls out address columns and lat and long from a geocoded file based on column mapping
# since this appends, you should either truncate the file or create a new blank one
def build_geocoded_address_catalog(
    input_file: str, output_file: str, column_mapping: Dict[str, str]
) -> None:
    # open output file in append mode
    with open(output_file, mode="a", newline="") as outfile:

        # create list of header fields and write header
        column_names = [c for c in column_mapping.keys()] + ["Latitude", "Longitude"]
        writer = csv.DictWriter(outfile, fieldnames=column_names, quoting=csv.QUOTE_ALL)
        writer.writeheader()

        # open input file and iterate over rows, writing out the selected fields
        with open(input_file, "r") as infile:
            reader = csv.DictReader(infile)
            for row in reader:
                columns = [column_mapping[col] for col in column_mapping.keys()]
                columns += ["Latitude", "Longitude"]
                selected_row = [row[col] for col in columns if col in row]
                writer.writerow(dict(zip(column_names, selected_row)))


# combines several address catalogs in a directory into a single file
def combine_address_catalogs(
    input_file_glob: str = "output/catalog/*.csv",
    output_file_path: pathlib.Path = output_directory
    / "catalog"
    / "full_address_catalog.csv",
) -> None:
    address_set = set()
    with output_file_path.open(mode="w", newline="") as output_file:
        fieldnames: Sequence[str] = []
        for input_file in glob.glob(input_file_glob):

            # ignore full catalog files, rename them if they should be included
            if "full_address_catalog" in input_file:
                continue

            # for each line in each file, try to add the address to the set
            with open(input_file, "r") as infile:
                catalog_reader = csv.DictReader(infile)

                for row in catalog_reader:
                    if catalog_reader.fieldnames is not None:
                        address_key = "|".join(
                            [row[k] for k in catalog_reader.fieldnames]
                        )
                        address_set.add(address_key)

        # iterate over address set and write out values that were geocoded
        catalog_writer = csv.DictWriter(
            output_file, fieldnames=fieldnames, quoting=csv.QUOTE_ALL
        )
        catalog_writer.writeheader()
        for key in address_set:
            row_dict = dict(zip(catalog_writer.fieldnames, key.split("|")))
            if float(row_dict["Latitude"]) == 0 and float(row_dict["Longitude"]) == 0:
                continue
            catalog_writer.writerow(row_dict)


# pulls unique addresses out of address catalog files and uses them to add geocoded data to input file
def extract_addresses_to_be_geocoded(
    address_catalog_glob: str,
    data_file_path: pathlib.Path,
    column_mapping: Dict[str, str],
) -> None:
    address_catalog_set = set()
    for f in glob.glob(address_catalog_glob):
        with open(f, "r") as address_catalog_file:
            catalog_reader = csv.DictReader(address_catalog_file)
            for row in catalog_reader:
                address_key = ",".join(
                    [row[k] for k in ["line1", "line2", "city", "state", "zipcode"]]
                )
                address_catalog_set.add(address_key)

    found_address_counter = 0
    with (
        data_file_path.open(mode="r") as data_file,
        open(f"output/{str(data_file_path)}_addresses_to_geocode.csv", "w") as out_file,
    ):
        data_reader = csv.DictReader(data_file)
        if data_reader.fieldnames is None:
            print("No fieldnames in data file")
            return

        data_reader_fieldnames = [field for field in data_reader.fieldnames]
        csv_writer = csv.DictWriter(out_file, fieldnames=data_reader_fieldnames)
        csv_writer.writeheader()
        for row in data_reader:
            address_key = ",".join([row[v] for k, v in column_mapping.items()])

            if address_key in address_catalog_set:
                found_address_counter += 1

            else:
                csv_writer.writerow(row)

    print(f"Found {found_address_counter} addresses cached")


# checks address cache collection for list of addresses
# returns a list of addresses missing from the address cache
def check_cache(
    collection: Collection[Dict[str, Any]],
    raw_addresses: List[RawAddress],
) -> Tuple[List[RawAddress], List[RawAddress]]:
    # save addresses in a dictionary to look up later
    raw_address_dict = {}
    for address in raw_addresses:
        raw_address_dict[address.to_hash()] = address

    # save address hashes in a set for easy lookup
    unique_address_hashes: Set[str] = set(raw_address_dict.keys())

    # find cached response for all hash keys in one request
    query = {"address_hash": {r"$in": list(unique_address_hashes)}}
    lookup_result: pymongo.cursor.Cursor[Any] = collection.find(query)

    # save cached hashes
    cached_hashes = set()
    for result in lookup_result:
        cached_hashes.add(result["address_hash"])

    # isolate missing hashes
    missing_hashes = set(unique_address_hashes - cached_hashes)

    missing_addresses_list = [
        raw_address_dict[address_hash] for address_hash in missing_hashes
    ]
    cached_addresses_list = [
        raw_address_dict[address_hash] for address_hash in cached_hashes
    ]

    return cached_addresses_list, missing_addresses_list


# saves vendor responses to address cache
def save_to_cache(
    _collection: Collection[Dict[str, Any]],
    vendor_responses: List[VendorResponse],
) -> None:
    requests = [
        UpdateOne(
            {"address_hash": vr.related_raw_address.to_hash()},
            {
                "$set": {
                    "vendor_std_address": vr.api_call_response,
                    "vendor_name": vr.vendor_name,
                    "response_version": vr.response_version,
                }
            },
            upsert=True,
        )
        for vr in vendor_responses
        if vr.related_raw_address
    ]
    if requests:
        _collection.bulk_write(requests=requests)


# uses a list of raw addresses to pull out the related vendor responses in a list
def filter_cached_addresses(
    cache_result: List[RawAddress], addresses: List[Dict[Any, Any]]
) -> List[Dict[Any, Any]]:
    uncached_address_hashes = set()
    for result in cache_result:
        uncached_address_hashes.add(result.to_hash())

    filtered_addresses = []
    for address in addresses:
        raw_address = RawAddress(
            address_id=address["RecordID"],
            line1=address["line1"],
            line2=address["line2"],
            city=address["city"],
            state=address["state"],
            zipcode=address["zipcode"],
        )
        if raw_address.to_hash() in uncached_address_hashes:
            filtered_addresses.append(address)

    return filtered_addresses


# takes an input catalog file and saves missing addresses to cache
def cache_addresses_from_file(
    catalog_file_path: pathlib.Path,
    address_mapping: Dict[str, str],
    connection_string: str,
    save_results_to_cache: bool = False,
    overwrite_cache: bool = False,
    batch_size: int = 10000,
) -> None:
    cache_miss_count = 0
    cache_hits_count = 0

    with catalog_file_path.open(mode="r") as fp:
        reader = csv.DictReader(fp)
        count = 0

        # perform in batches so the message to mongo doesn't grow too big
        for chunk in batch_process_csv_reader(reader, batch_size):
            raw_address_list: List[RawAddress] = []
            chunk_copy = []

            # iterate over rows and save a copy of each row (serves as vendor response)
            for row in chunk:
                row_copy = row.copy()
                raw_address_list.append(
                    RawAddress(
                        address_id=str(count),
                        line1=row[address_mapping["line1"]],
                        line2=row[address_mapping["line2"]],
                        city=row[address_mapping["city"]],
                        state=row[address_mapping["state"]],
                        zipcode=row[address_mapping["zipcode"]],
                    )
                )
                row_copy["RecordID"] = str(count)
                chunk_copy.append(row_copy)
                count += 1

            # get mongo client from connection_string
            mongo_client: MongoClient[Dict[str, Any]] = pymongo.MongoClient(
                connection_string
            )
            _db = mongo_client.get_database("helix_address_cache")
            __collection = _db["helix_address_cache"]

            # check cache for addresses in batch
            cache_hits, cache_misses = check_cache(
                raw_addresses=raw_address_list, collection=__collection
            )

            cache_hits_count += len(cache_hits)
            cache_miss_count += len(cache_misses)

            if save_results_to_cache:
                vendor_response = filter_cached_addresses(
                    cache_result=cache_misses, addresses=chunk_copy
                )
                # noinspection PyProtectedMember
                vendor_responses = StandardizingVendor()._to_vendor_response(
                    vendor_response=vendor_response,
                    raw_addresses=raw_address_list,
                    vendor_name="Geocodio",
                    response_version="1",
                )
                save_to_cache(
                    vendor_responses=vendor_responses, _collection=__collection
                )

                if overwrite_cache:
                    vendor_response = filter_cached_addresses(
                        cache_result=cache_hits, addresses=chunk_copy
                    )
                    # noinspection PyProtectedMember
                    vendor_responses = StandardizingVendor()._to_vendor_response(
                        vendor_response=vendor_response,
                        raw_addresses=raw_address_list,
                        vendor_name="Geocodio",
                        response_version="1",
                    )
                    save_to_cache(
                        vendor_responses=vendor_responses, _collection=__collection
                    )

    print(f"cache_hits={cache_hits_count} cache_misses={cache_miss_count}")



import json
from logging import Logger
from typing import Any, Dict, Iterator, List, Optional

from opensearchpy.helpers import bulk
from furl import furl
from opensearchpy.helpers.errors import BulkIndexError

from spark_pipeline_framework.transformers.elasticsearch_sender.v1.elasticsearch_result import (
    ElasticSearchResult,
)
from spark_pipeline_framework.utilities.elastic_search.elastic_search_connection import (
    ElasticSearchConnection,
)


def generate_es_bulk(
    input_data_json_str: List[str],
    operation: str,
    index: str,
    doc_id_prefix: Optional[str] = None,
) -> Iterator[Dict[str, Any]]:
    """this is how ES bulk function likes the input the be served
    https://elasticsearch-py.readthedocs.io/en/v7.12.0/helpers.html#bulk-helpers
    """
    for item in input_data_json_str:
        item_json = json.loads(item)
        try:
            doc_id = item_json["id"]
            if doc_id_prefix:
                doc_id = f"{doc_id_prefix}-{doc_id}"

            if operation != "index":
                wrapped_doc: Dict[str, Any] = {
                    "_op_type": operation,
                    "_index": index,
                    "_id": doc_id,
                    "doc": item_json,
                    "doc_as_upsert": True,
                }
                item_json = wrapped_doc

            if "_id" not in item_json.keys():
                item_json["_id"] = doc_id
        except KeyError:
            # just for local logging not usable for cluster env
            print(f"item without id >>> {item}")
            # raise e

        yield item_json


def send_json_bundle_to_elasticsearch(
    json_data_list: List[str],
    index: str,
    operation: str,
    logger: Logger,
    doc_id_prefix: Optional[str] = None,
) -> ElasticSearchResult:
    """
    Sends a list of json strings to the server
    (Runs on worker nodes so any logging will not show)

    :param json_data_list: list of json strings to send to server
    :param index: name of ElasticSearch index
    :param operation:
    :param logger: logger
    :param doc_id_prefix: a string to be prepended to the _id field for a document
    """
    es_connection = ElasticSearchConnection()
    es_client = es_connection.get_elastic_search_client()
    server_url = es_connection.get_elastic_search_host()
    payload: List[Dict[str, Any]] = list(
        generate_es_bulk(json_data_list, operation, index, doc_id_prefix)
    )
    success: int = 0
    failed: int = 0

    try:
        success, failed = bulk(
            client=es_client, actions=payload, index=index, stats_only=True
        )
    except BulkIndexError as e:
        for error in e.errors:
            logger.error(f"The following record failed to index: {error}")

    full_uri: furl = furl(server_url)
    full_uri /= index
    return ElasticSearchResult(
        url=full_uri.url,
        success=success,
        failed=failed,
        payload=list(payload),
        partition_index=0,
    )



from os import environ
from types import TracebackType
from typing import Optional, List, Dict, Any

from spark_pipeline_framework.event_loggers.event_logger import EventLogger
from spark_pipeline_framework.logger.log_level import LogLevel
from spark_pipeline_framework.logger.yarn_logger import get_logger


class ProgressLogger:
    def __init__(
        self,
        event_loggers: Optional[List[EventLogger]] = None,
    ) -> None:
        self.logger = get_logger(__name__)
        self.event_loggers: Optional[List[EventLogger]] = event_loggers
        system_log_level_text: Optional[str] = environ.get("LOGLEVEL")
        self.system_log_level: Optional[LogLevel] = (
            LogLevel.from_str(system_log_level_text)
            if system_log_level_text is not None
            else None
        )

    def __enter__(self) -> "ProgressLogger":
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.logger.debug("ENDING PARENT RUN")

    def log_metric(
        self,
        name: str,
        time_diff_in_minutes: float,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        self.logger.debug(f"{name}: {time_diff_in_minutes} min")

    def log_param(
        self, key: str, value: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        self.write_to_log(name=key, message=value)

    def log_params(
        self, params: Dict[str, Any], log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        pass

    # noinspection PyUnusedLocal
    def log_artifact(
        self, key: str, contents: str, folder_path: Optional[str] = None
    ) -> None:
        pass

    def write_to_log(
        self, name: str, message: str = "", log_level: LogLevel = LogLevel.INFO
    ) -> bool:
        if log_level == LogLevel.ERROR:
            self.logger.error(name + ": " + str(message))
        elif log_level == LogLevel.INFO:
            self.logger.info(name + ": " + str(message))
        else:
            self.logger.debug(name + ": " + str(message))
        return True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        self.log_artifact("_exception.txt", str(ex))
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_exception(
                    event_name=event_name, event_text=event_text, ex=ex
                )

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        self.logger.debug(event_format_string.format(event_name, current, total))
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_progress_event(
                            event_name=event_name,
                            current=current,
                            total=total,
                            event_format_string=event_format_string,
                            backoff=backoff,
                        )
        else:  # LOGLEVEL is lower than INFO
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_progress_event(
                        event_name=event_name,
                        current=current,
                        total=total,
                        event_format_string=event_format_string,
                        backoff=backoff,
                    )

    def log_event(
        self, event_name: str, event_text: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        self.write_to_log(name=event_name, message=event_text)
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_event(
                            event_name=event_name, event_text=event_text
                        )
        else:
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_event(event_name=event_name, event_text=event_text)

    @staticmethod
    def clean_experiments(*, tracking_uri: Optional[str]) -> None:
        pass



# mypy: ignore-errors

from datetime import datetime
from typing import Optional, Dict, Any, List, Tuple, Union

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.logger.yarn_logger import get_logger

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters

import pymongo
from spark_pipeline_framework.utilities.spark_data_frame_helpers import sc

import json
from bson.json_util import dumps


class ConnectHubDataReceiver(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view_name: str,
        conn_string: str,
        db_name: str,
        last_run_date: Optional[datetime] = datetime(1970, 1, 1),
        page_size: Optional[
            int
        ] = 0,  # setting to 0 ignores limit and will return all results
        query_parameters=Optional[Dict[str, Any]],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Transformer to call and receive data from ConnectHub (DocumentDB). It is specific to ConnectHub for now,
        but it should/will get refactored into a generic DocumentDbDataReceiver when time allows.

        :param view_name: name of the view to read the response into
        :param conn_string: the root URL
        :param db_name: the name of the database we need to query
        :param last_run_date: (Optional) the datetime which we should be querying from to get the latest updates. It is
        set to 1/1/1970 by default so that we grab all records without filtering (since there doesn't seem to be a good
        automated way to pass this in from pipeline runs at this point)
        :param page_size: (Optional) the number of documents we should be querying at a time, if we want paging.
        Setting to 0 (which is the default) will ignore paging and all documents (within the query parameters) will
        get pulled at once.
        :param query_parameters: Dict of key, value pairs which respects document query
        :param name: (Optional) name of transformer
        :param parameters: (Optional) parameters
        :param progress_logger: (Optional) progress logger
        """
        super(ConnectHubDataReceiver, self).__init__(
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
        )

        self.logger = get_logger(__name__)

        self.view_name: Param[str] = Param(self, "view_name", "")
        self._setDefault(view_name=view_name)

        self.page_size: Param[int] = Param(self, "page_size", "")
        self._setDefault(page_size=page_size)

        self.conn_string: Param[str] = Param(self, "conn_string", "")
        self._setDefault(conn_string=conn_string)

        self.last_run_date: Param[datetime] = Param(self, "last_run_date", "")
        self._setDefault(last_run_date=last_run_date)

        self.query_parameters: Param[Optional[Dict[str, Any]]] = Param(
            self, "query_parameters", ""
        )
        self._setDefault(query_parameters=query_parameters)

        self.db_name: Param[str] = Param(self, "db_name", "")
        self._setDefault(db_name=db_name)

    def _transform(self, df: DataFrame) -> DataFrame:
        view_name = self.get_view_name()
        conn_string = self.get_conn_string()
        db_name = self.get_db_name()
        query_parameters = self.get_query_parameters()
        client = pymongo.MongoClient(conn_string)  # type: ignore

        try:
            self.logger.info(f"connected server info: {client.server_info()}")

            integration_hub_db = client[db_name]
            client_connection = integration_hub_db.client_connection

            self.logger.info(
                f"db name: {db_name} and total # of documents in client_connection collection: "
                f"{client_connection.count_documents({})}"
            )

            converted_data: List[Any] = []
            last_seen = None
            while True:
                data, last_seen = self.pagination(
                    client_connection, query_parameters, last_seen
                )
                if not data:
                    break
                converted_data.extend(data)

            self.logger.info(
                f"the total number of documents about to be loaded into Spark: {len(converted_data)}"
            )

            df2 = df.sparkSession.read.json(
                sc(df).parallelize([json.dumps(r) for r in converted_data])
            )
            df2.createOrReplaceTempView(view_name)

        finally:
            client.close()

        return df

    def pagination(
        self, client: Any, query_parameters: Optional[Dict[str, Any]], last_seen  # type: ignore
    ) -> Tuple[Union[List[Any], None], Union[datetime, None]]:
        """
        Helper function to paginate results from ConnectHub. The built-in cursor.skip and cursor.limit methods
        have performance problems with large datasets, since all the skipped documents need to be read anyway.

        :param client: PyMongo client for connecting to 'client_connection' collection in 'integration_hub' db.
        :param query_parameters: Dict of key,value pairs which respects document query
        :param last_seen: The latest 'lastUpdatedOnDate' datetime for the last document that was pulled during the
        previous page, so that we can start paging from the next document after that. Note: if page_size is 0, all
        documents will be pulled at the same time without paging.
        :returns: A tuple consisting of the queried data (List) and the datetime representing the
        latest 'lastUpdatedOnDate' datetime for the last document that was pulled during this page.
        """
        page_size = self.get_page_size()
        last_run_date = self.get_last_run_date()

        self.logger.info(f"last_run_date: {last_run_date}; last_seen: {last_seen}")

        LAST_UPDATED_ON_DATE = "lastUpdatedOnDate"

        # When it is first page
        cursor = (
            client.find(
                {
                    LAST_UPDATED_ON_DATE: (
                        {"$gt": last_seen} if last_seen else {"$gte": last_run_date}
                    ),
                    **(query_parameters if query_parameters else {}),
                },
            )
            .sort(LAST_UPDATED_ON_DATE, pymongo.ASCENDING)
            .limit(page_size)
        )
        # Get the data
        list_cur = list(cursor)

        if not list_cur:
            # No documents left
            return None, None

        json_string = dumps(list_cur)
        data: List[Any] = json.loads(json_string)

        self.logger.info(f"returned # of documents: {len(data)}")

        # this appears to be unique, since it takes more than a ms between updates to generate this timestamp
        last_seen_string = data[-1][LAST_UPDATED_ON_DATE]["$date"]
        last_seen_date = datetime.strptime(last_seen_string, "%Y-%m-%dT%H:%M:%S.%fZ")

        self.logger.info(
            f"the {LAST_UPDATED_ON_DATE} of the last returned document: {last_seen_string}"
        )

        return data, last_seen_date

    def get_view_name(self) -> str:
        return self.getOrDefault(self.view_name)

    def get_page_size(self) -> int:
        return self.getOrDefault(self.page_size)

    def get_conn_string(self) -> str:
        return self.getOrDefault(self.conn_string)

    def get_last_run_date(self) -> datetime:
        return self.getOrDefault(self.last_run_date)

    def get_db_name(self) -> str:
        return self.getOrDefault(self.db_name)

    def get_query_parameters(self) -> Optional[Dict[str, Any]]:
        return self.getOrDefault(self.query_parameters)



import shutil
from os import path
from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from spark_fhir_schemas.r4.resources.patient import PatientSchema

from spark_pipeline_framework.transformers.fhir_reader.v1.fhir_reader import FhirReader


def test_fhir_reader_on_delta(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    test_files_dir: Path = data_dir.joinpath("test_files/patients")

    if path.isdir(data_dir.joinpath("temp")):
        shutil.rmtree(data_dir.joinpath("temp"))

    temp_folder = data_dir.joinpath("temp/")

    # write in delta format
    df = spark_session.read.json(str(test_files_dir))

    delta_file = temp_folder.joinpath("1.delta")
    df.write.format("delta").save(str(delta_file))

    # Act
    FhirReader(
        view="patients",
        schema=PatientSchema.get_schema(),
        file_path=delta_file,
        delta_lake_table="table",
    ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("patients")
    result_df.printSchema()
    result_df.show(truncate=False)

    assert result_df.count() == 2



import asyncio
import dataclasses
from datetime import datetime
from typing import (
    List,
    Dict,
    Any,
    Optional,
    AsyncGenerator,
    Iterable,
    Tuple,
    Generator,
)

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import struct, to_json
from pyspark.sql.types import StructType, StructField, StringType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_struct_column_to_struct_udf import (
    AsyncPandasStructColumnToStructColumnUDF,
)

from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


def test_async_pandas_struct_column_to_struct_column_udf(
    spark_session: SparkSession,
) -> None:
    print()
    df: DataFrame = spark_session.createDataFrame(
        [
            ("00100000000", "Qureshi"),
            ("00200000000", "Vidal"),
        ],
        ["id", "name"],
    )

    # add a json column
    df = df.withColumn("name_struct", to_json(struct("*")))

    print(f"Partition Count: {df.rdd.getNumPartitions()}")

    # Function to list partition IDs and column values
    def list_partition_ids_and_values(
        partition_index: int, iterator: Iterable[Any]
    ) -> Generator[Tuple[int, List[str]], None, None]:
        for row in iterator:
            yield partition_index, row["id"]

    partition_id_and_values: List[Tuple[int, List[str]]] = (
        df.rdd.mapPartitionsWithIndex(list_partition_ids_and_values).collect()
    )
    print(f"Partition IDs: {partition_id_and_values}")

    @dataclasses.dataclass
    class MyParameters:
        log_level: str = "INFO"

    # noinspection PyUnusedLocal
    async def test_async(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[MyParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        if parameters is not None and parameters.log_level == "DEBUG":
            spark_partition_information: SparkPartitionInformation = (
                SparkPartitionInformation.from_current_task_context(
                    chunk_index=run_context.chunk_index,
                )
            )
            logger = get_logger(__name__)
            ids = [input_value["id"] for input_value in input_values]
            message: str = f"In test_async"
            # Get the current time
            current_time = datetime.now()

            # Format the time to include hours, minutes, seconds, and milliseconds
            formatted_time = current_time.strftime("%H:%M:%S.%f")[:-3]
            print(
                f"{formatted_time}: "
                f"{message}"
                f" | Partition: {run_context.partition_index}"
                f" | Chunk: {run_context.chunk_index}"
                f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
                f" | Ids ({len(ids)}): {ids}"
                f" | {spark_partition_information}"
            )

        input_value: Dict[str, Any]
        for input_value in input_values:
            # simulate a delay
            await asyncio.sleep(0.1)
            yield {
                "id": input_value["id"],
                "name": input_value["name"] + "_processed",
            }

    result_df: DataFrame = df.withColumn(
        colName="processed_name",
        col=AsyncPandasStructColumnToStructColumnUDF(
            async_func=test_async,  # type: ignore[arg-type]
            parameters=MyParameters(),
            pandas_udf_parameters=AsyncPandasUdfParameters(max_chunk_size=2),
        ).get_pandas_udf(
            return_type=StructType(
                [StructField("id", StringType()), StructField("name", StringType())]
            ),
        )(
            df["name_struct"]
        ),
    )

    print("result_df")
    result_df.show(truncate=False)

    assert result_df.count() == 2
    assert result_df.select("processed_name.name").collect() == [
        ("Qureshi_processed",),
        ("Vidal_processed",),
    ]



import json
import uuid
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Callable, cast
from typing import (
    Iterable,
    AsyncGenerator,
    Iterator,
)

import pandas as pd

# noinspection PyPep8Naming
import pyspark.sql.functions as F
from pyspark import StorageLevel
from pyspark.errors import PythonException
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, from_json
from pyspark.sql.functions import explode
from pyspark.sql.types import (
    StringType,
    StructType,
    DataType,
)

from spark_pipeline_framework.logger.log_level import LogLevel
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_parameters import (
    FhirReceiverParameters,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_error import (
    GetBatchError,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_result import (
    GetBatchResult,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_dataframe_udf import (
    AsyncPandasDataFrameUDF,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_response_item import (
    FhirGetResponseItem,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_response_schema import (
    FhirGetResponseSchema,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_receiver_exception import (
    FhirReceiverException,
)
from spark_pipeline_framework.utilities.pretty_print import get_pretty_data_frame
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)
from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


class FhirReceiverProcessorSpark:
    """
    This class contains the methods to process the FHIR receiver in Spark
    """

    # noinspection PyUnusedLocal
    @staticmethod
    async def process_chunk(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[FhirReceiverParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Process a partition of data asynchronously

        :param run_context: run context
        :param input_values: input values
        :param parameters: parameters
        :param additional_parameters: additional parameters
        :return: output values
        """
        register()
        assert parameters
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        spark_partition_information: SparkPartitionInformation = (
            SparkPartitionInformation.from_current_task_context(
                chunk_index=run_context.chunk_index,
            )
        )
        # ids = [input_value["id"] for input_value in input_values]
        message: str = f"FhirReceiverProcessor.process_chunk"
        # Format the time to include hours, minutes, seconds, and milliseconds
        formatted_time = datetime.now().strftime("%H:%M:%S.%f")[:-3]
        formatted_message: str = (
            f"{formatted_time}: "
            + f"{message}"
            + f" | Partition: {run_context.partition_index}"
            + (f"/{parameters.total_partitions}" if parameters.total_partitions else "")
            + f" | Chunk: {run_context.chunk_index}"
            + f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
            + f" | {spark_partition_information}"
        )
        logger.info(formatted_message)

        try:
            r: Dict[str, Any]
            async for r in FhirReceiverProcessor.send_partition_request_to_server_async(
                partition_index=run_context.partition_index,
                parameters=parameters,
                rows=input_values,
            ):
                response_item: FhirGetResponseItem = FhirGetResponseItem.from_dict(r)
                # count += len(response_item.responses)
                logger.debug(
                    f"Received result"
                    f" | Partition: {run_context.partition_index}"
                    f" | Chunk: {run_context.chunk_index}"
                    f" | Status: {response_item.status_code}"
                    f" | Url: {response_item.url}"
                    f" | Count: {response_item.received}"
                )
                yield r
        except Exception as e:
            logger.error(
                f"Error processing partition {run_context.partition_index} chunk {run_context.chunk_index}: {str(e)}"
            )
            # if an exception is thrown then return an error for each row
            for _ in input_values:
                # count += 1
                yield {
                    FhirGetResponseSchema.partition_index: run_context.partition_index,
                    FhirGetResponseSchema.sent: 0,
                    FhirGetResponseSchema.received: 0,
                    FhirGetResponseSchema.url: parameters.server_url,
                    FhirGetResponseSchema.responses: [],
                    FhirGetResponseSchema.first: None,
                    FhirGetResponseSchema.last: None,
                    FhirGetResponseSchema.error_text: str(e),
                    FhirGetResponseSchema.status_code: 400,
                    FhirGetResponseSchema.request_id: None,
                    FhirGetResponseSchema.access_token: None,
                    FhirGetResponseSchema.extra_context_to_return: None,
                }

    @staticmethod
    def get_process_batch_function(
        *, parameters: FhirReceiverParameters
    ) -> Callable[[Iterable[pd.DataFrame]], Iterator[pd.DataFrame]]:
        """
        Returns a function that includes the passed parameters so that function
        can be used in a pandas_udf

        :param parameters: FhirSenderParameters
        :return: pandas_udf
        """

        return AsyncPandasDataFrameUDF(
            async_func=FhirReceiverProcessorSpark.process_chunk,  # type: ignore[arg-type]
            parameters=parameters,
            pandas_udf_parameters=AsyncPandasUdfParameters(
                max_chunk_size=parameters.batch_size or 100
            ),
        ).get_pandas_udf()

    @staticmethod
    async def get_batch_result_streaming_dataframe_async(
        *,
        df: DataFrame,
        schema: StructType,
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        parameters: FhirReceiverParameters,
        server_url: Optional[str],
        results_per_batch: Optional[int],
        limit: Optional[int] = None,
    ) -> DataFrame:
        """
        Converts the results from a batch streaming request to a DataFrame iteratively using
        batches of size `results_per_batch`

        :param df: DataFrame
        :param schema: StructType | AtomicType
        :param last_updated_after: Optional[datetime]
        :param last_updated_before: Optional[datetime]
        :param parameters: FhirReceiverParameters
        :param server_url: Optional[str]
        :param results_per_batch: int
        :param limit: int
        :return: DataFrame
        """
        return await AsyncHelper.async_generator_to_dataframe(
            df=df,
            async_gen=FhirReceiverProcessor.get_batch_result_streaming_async(
                last_updated_after=last_updated_after,
                last_updated_before=last_updated_before,
                parameters=parameters,
                server_url=server_url,
                limit=limit,
            ),
            schema=schema,
            results_per_batch=results_per_batch,
        )

    @staticmethod
    async def get_all_resources_async(
        *,
        df: DataFrame,
        parameters: FhirReceiverParameters,
        delta_lake_table: Optional[str],
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        batch_size: Optional[int],
        mode: str,
        file_path: Path | str | None,
        page_size: Optional[int],
        limit: Optional[int],
        progress_logger: Optional[ProgressLogger],
        view: Optional[str],
        error_view: Optional[str],
        logger: Logger,
        schema: Optional[Union[StructType, DataType]] = None,
    ) -> DataFrame:
        """
        This function gets all the resources from the FHIR server based on the resourceType and
        any additional query parameters and writes them to a file

        :param df: the data frame
        :param parameters: the parameters
        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param last_updated_after: only get records newer than this
        :param last_updated_before: only get records older than this
        :param batch_size: how many id rows to send to FHIR server in one call
        :param mode: if output files exist, should we overwrite or append
        :param file_path: where to store the FHIR resources retrieved
        :param page_size: use paging and get this many items in each page
        :param limit: maximum number of resources to get
        :param progress_logger: progress logger
        :param view: view to create
        :param error_view: view to create for errors
        :param logger: logger
        :param schema: the schema to apply after we receive the data
        :return: the data frame
        """
        file_format = "delta" if delta_lake_table else "text"

        if parameters.use_data_streaming:
            errors_df = (
                await FhirReceiverProcessorSpark.get_all_resources_streaming_async(
                    df=df,
                    batch_size=batch_size,
                    file_format=file_format,
                    file_path=file_path,
                    last_updated_after=last_updated_after,
                    last_updated_before=last_updated_before,
                    mode=mode,
                    parameters=parameters,
                    limit=limit,
                )
            )
        else:
            errors_df = (
                await FhirReceiverProcessorSpark.get_all_resources_non_streaming_async(
                    df=df,
                    file_format=file_format,
                    file_path=file_path,
                    last_updated_after=last_updated_after,
                    last_updated_before=last_updated_before,
                    limit=limit,
                    mode=mode,
                    page_size=page_size,
                    parameters=parameters,
                )
            )

        list_df = df.sparkSession.read.format(file_format).load(str(file_path))

        logger.info(f"Wrote FHIR data to {file_path}")

        if progress_logger:
            progress_logger.log_event(
                event_name="Finished receiving FHIR",
                event_text=json.dumps(
                    {
                        "message": f"Wrote {list_df.count()} FHIR {parameters.resource_type} resources to "
                        + f"{file_path} (query)",
                        "count": list_df.count(),
                        "resourceType": parameters.resource_type,
                        "path": str(file_path),
                    },
                    default=str,
                ),
            )

        if view:
            if schema:
                resource_df = (
                    df.sparkSession.read.schema(schema)  # type: ignore
                    .format("json")
                    .load(str(file_path))
                )
            else:
                resource_df = df.sparkSession.read.format("json").load(str(file_path))
            resource_df.createOrReplaceTempView(view)
        if error_view:
            errors_df.createOrReplaceTempView(error_view)
            if progress_logger and not spark_is_data_frame_empty(errors_df):
                progress_logger.log_event(
                    event_name="Errors receiving FHIR",
                    event_text=get_pretty_data_frame(
                        df=errors_df,
                        limit=100,
                        name="Errors Receiving FHIR",
                    ),
                    log_level=LogLevel.INFO,
                )

        return df

    @staticmethod
    async def get_all_resources_non_streaming_async(
        *,
        df: DataFrame,
        file_format: str,
        file_path: Path | str | None,
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        limit: Optional[int],
        mode: str,
        page_size: Optional[int],
        parameters: FhirReceiverParameters,
    ) -> DataFrame:
        """
        This function gets all the resources from the FHIR server based on the resourceType and
        any additional query parameters and writes them to a file.  This function does not use data streaming


        :param df: the data frame
        :param file_format: the file format
        :param file_path: where to store the FHIR resources retrieved
        :param last_updated_after: only get records newer than this
        :param last_updated_before: only get records older than this
        :param limit: maximum number of resources to get
        :param mode: if output files exist, should we overwrite or append
        :param page_size: use paging and get this many items in each page
        :param parameters: the parameters
        :return: the data frame
        """
        resources: List[str] = []
        errors: List[GetBatchError] = []
        async for result1 in FhirReceiverProcessor.get_batch_results_paging_async(
            page_size=page_size,
            limit=limit,
            server_url=parameters.server_url,
            parameters=parameters,
            last_updated_after=last_updated_after,
            last_updated_before=last_updated_before,
        ):
            resources.extend(result1.resources)
            errors.extend(result1.errors)
            if limit and 0 < limit <= len(resources):
                break
        list_df = df.sparkSession.createDataFrame(resources, schema=StringType())
        errors_df = (
            df.sparkSession.createDataFrame(  # type:ignore[type-var]
                [e.to_dict() for e in errors],
                schema=GetBatchError.get_schema(),
            )
            if errors
            else df.sparkSession.createDataFrame([], schema=StringType())
        )
        list_df.write.format(file_format).mode(mode).save(str(file_path))
        return errors_df

    @staticmethod
    async def get_all_resources_streaming_async(
        *,
        df: DataFrame,
        batch_size: Optional[int],
        file_format: str,
        file_path: Path | str | None,
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        mode: str,
        parameters: FhirReceiverParameters,
        limit: Optional[int] = None,
    ) -> DataFrame:
        """
        Get all resources from the FHIR server based on the resourceType and any additional query parameters
        and write them to a file.  This function uses data streaming.

        :param df: the data frame
        :param batch_size: how many id rows to send to FHIR server in one call
        :param file_format: the file format
        :param file_path: where to store the FHIR resources retrieved
        :param last_updated_after: only get records newer than this
        :param last_updated_before: only get records older than this
        :param mode: if output files exist, should we overwrite or append
        :param parameters: the parameters
        :param limit: maximum number of resources to get
        :return: the data frame
        """
        list_df: DataFrame = (
            await FhirReceiverProcessorSpark.get_batch_result_streaming_dataframe_async(
                df=df,
                server_url=parameters.server_url,
                parameters=parameters,
                last_updated_after=last_updated_after,
                last_updated_before=last_updated_before,
                schema=GetBatchResult.get_schema(),
                results_per_batch=batch_size,
                limit=limit,
            )
        )
        resource_df = list_df.select(explode(col("resources")).alias("resource"))
        errors_df = list_df.select(explode(col("errors")).alias("resource")).select(
            "resource.*"
        )
        resource_df.write.format(file_format).mode(mode).save(str(file_path))
        return errors_df

    @staticmethod
    async def get_resources_by_id_view_async(
        *,
        df: DataFrame,
        id_view: str,
        parameters: FhirReceiverParameters,
        run_synchronously: Optional[bool],
        checkpoint_path: Path | str | None,
        progress_logger: Optional[ProgressLogger],
        delta_lake_table: Optional[str],
        cache_storage_level: Optional[StorageLevel],
        error_view: Optional[str],
        name: Optional[str],
        expand_fhir_bundle: Optional[bool],
        error_on_result_count: Optional[bool],
        verify_counts_match: Optional[bool],
        file_path: Union[Path, str],
        schema: Optional[Union[StructType, DataType]],
        mode: str,
        view: Optional[str],
        logger: Logger,
        loop_id: Optional[str] = None,
    ) -> DataFrame:
        """
        Gets the resources by id from id_view and writes them to a file

        :param df: the data frame
        :param id_view: the view that contains the ids to retrieve
        :param parameters: the parameters
        :param run_synchronously: run synchronously
        :param checkpoint_path: where to store the checkpoint
        :param progress_logger: progress logger
        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param cache_storage_level: how to store the cache
        :param error_view: view to create for errors
        :param name: name of transformer
        :param expand_fhir_bundle: expand the FHIR bundle
        :param error_on_result_count: throw exception when the response from FHIR count does not match the request count
        :param verify_counts_match: verify counts match
        :param file_path: where to store the FHIR resources retrieved
        :param schema: the schema to apply after we receive the data
        :param mode: if output files exist, should we overwrite or append
        :param view: view to create
        :param logger: the logger
        :param loop_id: the loop id
        :return: the data frame
        """
        id_df: DataFrame = df.sparkSession.table(id_view)
        if spark_is_data_frame_empty(df=id_df):
            # nothing to do
            return df

        # if id is not in the columns, but we have a value column with StructType then select the id in there
        # This means the data frame contains the id in a nested structure
        if not "id" in id_df.columns and "value" in id_df.columns:
            if isinstance(id_df.schema["value"].dataType, StructType):
                id_df = id_df.select(
                    F.col("value.resourceType").alias("resourceType"),
                    F.col("value.id").alias("id"),
                )

        assert "id" in id_df.columns

        has_token_col: bool = "token" in id_df.columns

        parameters.has_token_col = has_token_col

        if parameters.limit and parameters.limit > 0:
            id_df = id_df.limit(parameters.limit)

        row_count: int = id_df.count()

        logger.info(
            f"----- Total {row_count} rows to request from {parameters.server_url or ''}/{parameters.resource_type}  -----"
        )
        # see if we need to partition the incoming dataframe
        total_partitions: int = id_df.rdd.getNumPartitions()

        parameters.total_partitions = total_partitions

        logger.info(
            f"----- Total Batches: {total_partitions} for {parameters.server_url or ''}/{parameters.resource_type}  -----"
        )

        result_with_counts_and_responses: DataFrame
        if run_synchronously:
            result_with_counts_and_responses = await FhirReceiverProcessorSpark.get_resources_by_id_using_driver_node_async(
                df=df,
                id_df=id_df,
                parameters=parameters,
            )
        else:
            result_with_counts_and_responses = (
                await FhirReceiverProcessorSpark.get_resources_by_id_async(
                    has_token_col=has_token_col,
                    id_df=id_df,
                    parameters=parameters,
                )
            )

        try:
            # Now write to checkpoint if requested
            if checkpoint_path:
                result_with_counts_and_responses = await FhirReceiverProcessorSpark.write_to_checkpoint_async(
                    checkpoint_path=checkpoint_path,
                    delta_lake_table=delta_lake_table,
                    df=result_with_counts_and_responses,
                    parameters=parameters,
                    progress_logger=progress_logger,
                    result_with_counts_and_responses=result_with_counts_and_responses,
                    loop_id=loop_id,
                    name=name,
                )
            else:
                result_with_counts_and_responses = await FhirReceiverProcessorSpark.write_to_cache_async(
                    cache_storage_level=cache_storage_level,
                    result_with_counts_and_responses=result_with_counts_and_responses,
                )
            # don't need the large responses column to figure out if we had errors or not
            result_with_counts: DataFrame = result_with_counts_and_responses.select(
                FhirGetResponseSchema.partition_index,
                FhirGetResponseSchema.sent,
                FhirGetResponseSchema.received,
                FhirGetResponseSchema.first,
                FhirGetResponseSchema.last,
                FhirGetResponseSchema.error_text,
                FhirGetResponseSchema.url,
                FhirGetResponseSchema.status_code,
                FhirGetResponseSchema.request_id,
            )
            result_with_counts = result_with_counts.cache()
            # find results that have a status code not in the ignore_status_codes
            results_with_counts_errored = result_with_counts.where(
                (
                    (col(FhirGetResponseSchema.status_code).isNotNull())
                    & (
                        ~col(FhirGetResponseSchema.status_code).isin(
                            parameters.ignore_status_codes
                        )
                    )
                )
                | (col(FhirGetResponseSchema.error_text).isNotNull())
            )
            count_bad_requests: int = results_with_counts_errored.count()

            if count_bad_requests > 0:
                result_with_counts = (
                    await FhirReceiverProcessorSpark.handle_error_async(
                        error_view=error_view,
                        parameters=parameters,
                        progress_logger=progress_logger,
                        result_with_counts=result_with_counts,
                        results_with_counts_errored=results_with_counts_errored,
                    )
                )
        except PythonException as e:
            # noinspection PyUnresolvedReferences
            if hasattr(e, "desc") and "pyarrow.lib.ArrowTypeError" in e.desc:
                raise FriendlySparkException(
                    exception=e,
                    message="Exception converting data to Arrow format."
                    + f" This is usually because the return data did not match the specified schema.",
                    stage_name=name,
                )
            else:
                raise
        except Exception as e:
            raise FriendlySparkException(exception=e, stage_name=name)

        result_df: DataFrame = (
            await FhirReceiverProcessorSpark.expand_bundle_async(
                error_on_result_count=error_on_result_count,
                parameters=parameters,
                result_with_counts=result_with_counts,
                verify_counts_match=verify_counts_match,
                result_with_counts_and_responses=result_with_counts_and_responses,
                logger=logger,
            )
            if expand_fhir_bundle
            else result_with_counts_and_responses.select(
                col(FhirGetResponseSchema.responses)[0].alias("bundle")
            )
        )

        # result_df.printSchema()
        # TODO: remove any OperationOutcomes.  Some FHIR servers like to return these

        logger.info(
            f"Executing requests and writing FHIR {parameters.resource_type} resources to {file_path}..."
        )
        result_df = await FhirReceiverProcessorSpark.write_to_disk_async(
            delta_lake_table=delta_lake_table,
            df=df,
            file_path=file_path,
            mode=mode,
            result_df=result_df,
            schema=schema,
        )

        logger.info(
            f"Received {result_df.count()} FHIR {parameters.resource_type} resources."
        )
        logger.info(
            f"Reading from disk and counting rows for {parameters.resource_type}..."
        )
        file_row_count: int = result_df.count()
        logger.info(
            f"Wrote {file_row_count} FHIR {parameters.resource_type} resources to {file_path}"
        )
        if progress_logger:
            progress_logger.log_event(
                event_name="Finished receiving FHIR",
                event_text=json.dumps(
                    {
                        "message": f"Wrote {file_row_count} FHIR {parameters.resource_type} resources "
                        + f"to {file_path} (id view)",
                        "count": file_row_count,
                        "resourceType": parameters.resource_type,
                        "path": str(file_path),
                    },
                    default=str,
                ),
            )
        if view:
            result_df.createOrReplaceTempView(view)
        return df

    @staticmethod
    async def expand_bundle_async(
        *,
        error_on_result_count: Optional[bool],
        parameters: FhirReceiverParameters,
        result_with_counts: DataFrame,
        verify_counts_match: Optional[bool],
        result_with_counts_and_responses: DataFrame,
        logger: Logger,
    ) -> DataFrame:
        """
        Expand the FHIR bundle to return the included resources

        :param error_on_result_count: throw exception when the response from FHIR count does not match the request count
        :param parameters: the parameters
        :param result_with_counts: the result with counts
        :param verify_counts_match: verify counts match
        :param result_with_counts_and_responses: the result with counts and responses
        :param logger: the logger
        :return: the data frame
        """
        # noinspection PyUnresolvedReferences
        count_sent_df: DataFrame = result_with_counts.agg(
            F.sum(FhirGetResponseSchema.sent).alias(FhirGetResponseSchema.sent)
        )
        # count_sent_df.show()
        # noinspection PyUnresolvedReferences
        count_received_df: DataFrame = result_with_counts.agg(
            F.sum(FhirGetResponseSchema.received).alias(FhirGetResponseSchema.received)
        )
        # count_received_df.show()
        count_sent: Optional[int] = count_sent_df.collect()[0][0]
        count_received: Optional[int] = count_received_df.collect()[0][0]
        if (
            (count_sent is not None and count_received is not None)
            and (count_sent > count_received)
            and error_on_result_count
            and verify_counts_match
        ):
            result_with_counts.where(
                col(FhirGetResponseSchema.sent) != col(FhirGetResponseSchema.received)
            ).select(
                FhirGetResponseSchema.partition_index,
                FhirGetResponseSchema.sent,
                FhirGetResponseSchema.received,
                FhirGetResponseSchema.error_text,
                FhirGetResponseSchema.url,
                FhirGetResponseSchema.request_id,
            ).show(
                truncate=False, n=1000000
            )
            first_url = (
                result_with_counts.select(FhirGetResponseSchema.url)
                .limit(1)
                .collect()[0][0]
            )
            first_error_status_code = (
                result_with_counts.select(FhirGetResponseSchema.status_code)
                .limit(1)
                .collect()[0][0]
            )
            first_request_id = (
                result_with_counts.select(FhirGetResponseSchema.request_id)
                .limit(1)
                .collect()[0][0]
            )
            raise FhirReceiverException(
                url=first_url,
                response_text=None,
                response_status_code=first_error_status_code,
                message=f"Sent ({count_sent}) and Received ({count_received}) counts did not match",
                json_data="",
                request_id=first_request_id,
            )
        elif count_sent == count_received:
            logger.info(
                f"Sent ({count_sent}) and Received ({count_received}) counts matched "
                + f"for {parameters.server_url or ''}/{parameters.resource_type}"
            )
        else:
            logger.info(
                f"Sent ({count_sent}) and Received ({count_received}) for "
                + f"{parameters.server_url or ''}/{parameters.resource_type}"
            )

        return result_with_counts_and_responses.select(
            explode(col(FhirGetResponseSchema.responses))
        )

    @staticmethod
    async def handle_error_async(
        *,
        error_view: Optional[str],
        parameters: FhirReceiverParameters,
        progress_logger: Optional[ProgressLogger],
        result_with_counts: DataFrame,
        results_with_counts_errored: DataFrame,
    ) -> DataFrame:
        """
        Display errors and filter out bad records

        :param error_view: view to create for errors
        :param parameters: the parameters
        :param progress_logger: progress logger
        :param result_with_counts: the result with counts
        :param results_with_counts_errored: the result with counts and errors
        :return: the data frame
        """
        result_with_counts.select(
            FhirGetResponseSchema.partition_index,
            FhirGetResponseSchema.sent,
            FhirGetResponseSchema.received,
            FhirGetResponseSchema.error_text,
            FhirGetResponseSchema.status_code,
            FhirGetResponseSchema.request_id,
        ).show(truncate=False, n=1000)
        results_with_counts_errored.select(
            FhirGetResponseSchema.partition_index,
            FhirGetResponseSchema.sent,
            FhirGetResponseSchema.received,
            FhirGetResponseSchema.error_text,
            FhirGetResponseSchema.url,
            FhirGetResponseSchema.status_code,
            FhirGetResponseSchema.request_id,
        ).show(truncate=False, n=1000)
        first_error: str = (
            results_with_counts_errored.select(FhirGetResponseSchema.error_text)
            .limit(1)
            .collect()[0][0]
        )
        first_url: str = (
            results_with_counts_errored.select(FhirGetResponseSchema.url)
            .limit(1)
            .collect()[0][0]
        )
        first_error_status_code: Optional[int] = (
            results_with_counts_errored.select(FhirGetResponseSchema.status_code)
            .limit(1)
            .collect()[0][0]
        )
        first_request_id: Optional[str] = (
            results_with_counts_errored.select(FhirGetResponseSchema.request_id)
            .limit(1)
            .collect()[0][0]
        )
        if error_view:
            errors_df = results_with_counts_errored.select(
                FhirGetResponseSchema.url,
                FhirGetResponseSchema.error_text,
                FhirGetResponseSchema.status_code,
                FhirGetResponseSchema.request_id,
            )
            errors_df.createOrReplaceTempView(error_view)
            if progress_logger and not spark_is_data_frame_empty(errors_df):
                progress_logger.log_event(
                    event_name="Errors receiving FHIR",
                    event_text=get_pretty_data_frame(
                        df=errors_df,
                        limit=100,
                        name="Errors Receiving FHIR",
                    ),
                    log_level=LogLevel.INFO,
                )
            # filter out bad records
            result_with_counts = result_with_counts.where(
                ~(
                    (col(FhirGetResponseSchema.status_code).isNotNull())
                    & (
                        ~col(FhirGetResponseSchema.status_code).isin(
                            parameters.ignore_status_codes
                        )
                    )
                )
                | (col(FhirGetResponseSchema.error_text).isNotNull())
            )
        else:
            raise FhirReceiverException(
                url=first_url,
                response_text=first_error,
                response_status_code=first_error_status_code,
                message="Error receiving FHIR",
                json_data=first_error,
                request_id=first_request_id,
            )
        return result_with_counts

    @staticmethod
    async def write_to_cache_async(
        *,
        cache_storage_level: Optional[StorageLevel],
        result_with_counts_and_responses: DataFrame,
    ) -> DataFrame:
        """
        Write the results to cache depending on the storage level

        :param cache_storage_level: how to store the cache
        :param result_with_counts_and_responses: the result with counts and responses
        :return: the data frame
        """
        if cache_storage_level is None:
            result_with_counts_and_responses = result_with_counts_and_responses.cache()
        else:
            result_with_counts_and_responses = result_with_counts_and_responses.persist(
                storageLevel=cache_storage_level
            )
        return result_with_counts_and_responses

    @staticmethod
    async def write_to_checkpoint_async(
        *,
        checkpoint_path: Path | str,
        delta_lake_table: Optional[str],
        df: DataFrame,
        parameters: FhirReceiverParameters,
        progress_logger: Optional[ProgressLogger],
        result_with_counts_and_responses: DataFrame,
        loop_id: Optional[str],
        name: Optional[str],
    ) -> DataFrame:
        """
        Writes the data frame to the checkpoint path

        :param checkpoint_path: where to store the checkpoint
        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param df: the data frame
        :param parameters: the parameters
        :param progress_logger: progress logger
        :param result_with_counts_and_responses: the result with counts and responses
        :param loop_id: the loop id
        :param name: the name of the transformer
        :return: the data frame
        """
        if callable(checkpoint_path):
            checkpoint_path = checkpoint_path(loop_id)
        checkpoint_file = f"{checkpoint_path}/{parameters.resource_type}/{uuid.uuid4()}"
        if progress_logger:
            progress_logger.write_to_log(
                name or "FhirReceiverProcessor",
                f"Writing checkpoint to {checkpoint_file}",
            )
        checkpoint_file_format = "delta" if delta_lake_table else "parquet"
        result_with_counts_and_responses.write.format(checkpoint_file_format).save(
            checkpoint_file
        )
        result_with_counts_and_responses = df.sparkSession.read.format(
            checkpoint_file_format
        ).load(checkpoint_file)
        return result_with_counts_and_responses

    @staticmethod
    async def write_to_disk_async(
        *,
        delta_lake_table: Optional[str],
        df: DataFrame,
        file_path: Union[Path, str],
        mode: str,
        result_df: DataFrame,
        schema: Optional[Union[StructType, DataType]],
    ) -> DataFrame:
        """
        Writes the data frame to disk

        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param df: the data frame
        :param file_path: where to store the FHIR resources retrieved
        :param mode: if output files exist, should we overwrite or append
        :param result_df: the result data frame
        :param schema: the schema to apply after we receive the data
        :return: dataframe
        """
        if delta_lake_table:
            if schema:
                result_df = result_df.select(
                    from_json(col("col"), schema=cast(StructType, schema)).alias(
                        "resource"
                    )
                )
                result_df = result_df.selectExpr("resource.*")
            result_df.write.format("delta").mode(mode).save(str(file_path))
            result_df = df.sparkSession.read.format("delta").load(str(file_path))
        else:
            result_df.write.format("text").mode(mode).save(str(file_path))
            result_df = df.sparkSession.read.format("text").load(str(file_path))
        return result_df

    @staticmethod
    async def get_resources_by_id_async(
        *,
        has_token_col: Optional[bool],
        id_df: DataFrame,
        parameters: FhirReceiverParameters,
    ) -> DataFrame:
        """
        Get resources by id using data streaming

        :param has_token_col: whether we have a token column
        :param id_df: the data frame containing the ids to use for retrieving resources
        :param parameters: the parameters
        :return: the data frame
        """
        # ---- Now process all the results ----
        if has_token_col and not parameters.server_url:
            assert parameters.slug_column
            assert parameters.url_column
            assert all(
                [
                    c
                    for c in [
                        parameters.url_column,
                        parameters.slug_column,
                        "resourceType",
                    ]
                    if [c in id_df.columns]
                ]
            )
        # use mapInPandas
        result_with_counts_and_responses = id_df.mapInPandas(
            FhirReceiverProcessorSpark.get_process_batch_function(
                parameters=parameters
            ),
            schema=FhirGetResponseSchema.get_schema(),
        )
        return result_with_counts_and_responses

    @staticmethod
    async def get_resources_by_id_using_driver_node_async(
        *,
        df: DataFrame,
        id_df: DataFrame,
        parameters: FhirReceiverParameters,
    ) -> DataFrame:
        """
        Gets the resources by id from id_df and writes them to a file.  Does not use streaming.

        :param df: the data frame
        :param id_df: the data frame containing the ids to use for retrieving resources
        :param parameters: the parameters
        :return: the data frame
        """
        id_rows: List[Dict[str, Any]] = [
            r.asDict(recursive=True) for r in id_df.collect()
        ]
        result_rows: List[Dict[str, Any]] = await AsyncHelper.collect_items(
            FhirReceiverProcessor.send_partition_request_to_server_async(
                partition_index=0,
                rows=id_rows,
                parameters=parameters,
            )
        )
        response_schema = FhirGetResponseSchema.get_schema()
        result_with_counts_and_responses = (
            df.sparkSession.createDataFrame(  # type:ignore[type-var]
                result_rows, schema=response_schema
            )
        )
        return result_with_counts_and_responses



from typing import Dict, Any, Optional

from pydantic import BaseModel

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


class MelissaStandardizingVendorApiResponse(BaseModel, BaseVendorApiResponse):

    class Config:
        extra = "ignore"

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    RecordID: str
    FormattedAddress: Optional[str] = None
    Locality: Optional[str] = None
    AdministrativeArea: Optional[str] = None
    SubAdministrativeArea: Optional[str] = None
    PostalCode: Optional[str] = None
    CountryISO3166_1_Alpha2: Optional[str] = None
    Latitude: Optional[str] = None
    Longitude: Optional[str] = None

    def to_standardized_address(self, *, address_id: str) -> StandardizedAddress:
        return StandardizedAddress(
            address_id=address_id,
            line1=next(iter((self.FormattedAddress or "").split(";"))),
            line2=None,
            county=self.SubAdministrativeArea,
            city=self.Locality,
            state=self.AdministrativeArea,
            zipcode=self.PostalCode,
            country=self.CountryISO3166_1_Alpha2,
            latitude=str(self.Latitude) if self.Latitude else None,
            longitude=str(self.Longitude) if self.Longitude else None,
            formatted_address=self.FormattedAddress,
            standardize_vendor="melissa",
        )

    @classmethod
    def from_standardized_address(
        cls, standardized_address: StandardizedAddress
    ) -> "MelissaStandardizingVendorApiResponse":
        return cls(
            RecordID=standardized_address.address_id,
            FormattedAddress=standardized_address.formatted_address,
            Locality=standardized_address.city,
            AdministrativeArea=standardized_address.state,
            SubAdministrativeArea=standardized_address.county,
            PostalCode=standardized_address.zipcode,
            CountryISO3166_1_Alpha2=standardized_address.country,
            Latitude=standardized_address.latitude,
            Longitude=standardized_address.longitude,
        )
